#pragma once

#ifndef _CoordinateTools_H
#define _CoordinateTools_H

#include <math.h>
#include <algorithm>
#include <limits>
#include <string>
#include <vector>

struct CoordinatePoint{
	double x[3];
	CoordinatePoint(const double xx, const double yy, const double zz){
		x[0]=xx; x[1]=yy; x[2]=zz;
	};
	inline CoordinatePoint& operator+=(const CoordinatePoint& point){
		x[0]+=point.x[0]; x[1]+=point.x[1]; x[2]+=point.x[2]; 
		return *this;
	}
	inline CoordinatePoint& operator-=(const CoordinatePoint& point){
		x[0]-=point.x[0]; x[1]-=point.x[1]; x[2]-=point.x[2]; 
		return *this;
	}
	inline CoordinatePoint& operator*=(const double scale){
		x[0]*=scale; x[1]*=scale; x[2]*=scale; 
		return *this;
	}
	inline CoordinatePoint& operator/=(const double scale){
		return (*this)*=(1.0/scale);
	}
	inline double operator*(const CoordinatePoint& pt){
		return (x[0]*pt.x[0]+x[1]*pt.x[1]+x[2]*pt.x[2]);
	}
};
inline double Distance2(const CoordinatePoint& p1, const CoordinatePoint& p2){
	double t0=(p1.x[0]-p2.x[0]), t1=(p1.x[1]-p2.x[1]), t2=(p1.x[2]-p2.x[2]);
	return (t0*t0+t1*t1+t2*t2);
};

inline double Distance(const CoordinatePoint& p1, const CoordinatePoint& p2){
	return sqrt(Distance2(p1, p2));
};

inline CoordinatePoint operator-(const CoordinatePoint& p1, const CoordinatePoint& p2){
	double t0=(p1.x[0]-p2.x[0]), t1=(p1.x[1]-p2.x[1]), t2=(p1.x[2]-p2.x[2]);
	return CoordinatePoint(t0, t1, t2);
};

inline CoordinatePoint operator+(const CoordinatePoint& p1, const CoordinatePoint& p2){
	double t0=(p1.x[0]+p2.x[0]), t1=(p1.x[1]+p2.x[1]), t2=(p1.x[2]+p2.x[2]);
	return CoordinatePoint(t0, t1, t2);
};
inline CoordinatePoint operator*(const double scale, const CoordinatePoint& p2){
	CoordinatePoint p(p2);
	p*=scale;
	return p;
};

struct CoordinateSet{
	std::vector<CoordinatePoint> points;
	CoordinateSet(){
		points.clear();
	};

	CoordinateSet(const std::vector<CoordinatePoint> pts){
		points=pts;
	};

	CoordinateSet& operator+=(const CoordinatePoint& point){
		points.push_back(point);
		return *this;
	};
	CoordinateSet& operator+=(const CoordinateSet& pointSet){
		for (unsigned i=0; i<pointSet.points.size(); i++){
			points.push_back(pointSet.points[i]);
		}
		return *this;
	};
};

struct Rotation{
	double U[3][3];
	Rotation(){
		for (unsigned i=0; i<3; i++)
			for (unsigned j=0; j<3; j++)
				U[i][j]= ((i==j) ? 1.0 : 0.0 );
	};
	Rotation(const double in[3][3]){
		for (unsigned i=0; i<3; i++)
			for (unsigned j=0; j<3; j++)
				U[i][j]= in[i][j];
	};
	Rotation& operator=(const double in[3][3]){
		for (unsigned i=0; i<3; i++)
			for (unsigned j=0; j<3; j++)
				U[i][j]= in[i][j];
		return *this;
	};

	inline void Rotate(CoordinatePoint& input) const{

		CoordinatePoint output(0.0, 0.0, 0.0);
		for (unsigned i=0; i<3; i++)
			for (unsigned j=0; j<3; j++)
				output.x[i]+=U[i][j]*input.x[j];

		input=output;

	};
	inline void Rotate(CoordinateSet& input) const{
		for (unsigned iPoint=0; iPoint< input.points.size(); iPoint++)
			Rotate(input.points[iPoint]);
	};
};

#endif 
