#pragma once

#ifndef _MathCommon_H
#define _MathCommon_H

#include <math.h>
#include <algorithm>
#include <limits>
#include <string>
#include <vector>

#include <dlib/array.h>
#include <dlib/matrix.h>

#include "alglib/src/ap.h"
#include "alglib/src/integration.h"

const static double exp_limit=log(std::numeric_limits<double>::epsilon());
const static double mini_double=std::numeric_limits<double>::min();
const unsigned numberQuadradureWeightPoints=48;

inline bool equal(double a, double b, double delta)
{
    if (fabs(a*b)<mini_double)
		return fabs(a - b) <= delta;
	else
		return fabs(a - b) <= ( (fabs(a) < fabs(b) ? fabs(b) : fabs(a)) * delta);
}

inline bool equal(double a, double b){
	return equal(a, b, 1e-10);
}

typedef dlib::array<double> dlib_parameters;
typedef dlib::matrix<double,0,1> dlib_ParameterSet;

typedef std::vector<double> OneDDataSet;
typedef std::vector<OneDDataSet> OneDDataSetCollection;

typedef std::vector<double> NDData;
typedef std::vector<NDData> NDDataSet;
typedef std::vector<NDDataSet> NDDataSetCollection;

inline bool operator==(const NDDataSet& lhs, const NDDataSet& rhs){
	bool result=lhs.size()==rhs.size();
	if (result){
		unsigned dim=lhs.size();
		for (unsigned idim=0; idim<dim; idim++){
			result=( lhs[idim].size()==rhs[idim].size());
			if (!result) break;
			else {
				for (unsigned i=0; i<lhs[idim].size(); i++){
					result=(lhs[idim][i]==rhs[idim][i]);
					if (!result) break;
				}
			}
			if (!result) break;
		}
	}
	return result;
}

typedef std::vector<int> NDInt;
typedef std::vector<NDInt> NDIntSet;

struct GuassianDistribution{
	double mean;
	double variance;
	GuassianDistribution& operator=(const GuassianDistribution& other){
		mean=other.mean;
		variance=other.variance;
		return *this;
	};
};
typedef std::vector<GuassianDistribution> GuassianDistributionSet;

struct DataRange{
	double min;
	double max;
};
typedef std::vector<DataRange> DataRangeSet;

struct SplineNode{
	unsigned index[2];
	double position[2];
	double previous[2];
	double x[4][4];
	double minDist;
	int count;
	// 1D case
	SplineNode(const unsigned i, const double pos){
		index[0]=i;		
		index[1]=0;
		position[0]=pos;
		std::fill_n(x[0], 4, 0.0);
		minDist=0.0;
		count=0;
	};
	// 2D case
	SplineNode(const unsigned ix, const double posX, const unsigned iy, const double posY){
		index[0]=ix; index[1]=iy;	
		position[0]=posX; position[1]=posY;
		for (unsigned i=0; i<4; i++) std::fill_n(x[i], 4, 0.0);
		minDist=0.0;
		count=0;
	};
	bool operator<(const SplineNode& other) const {	return (this->position[0] < other.position[0]); };
};

struct QuadraturePoint{
	double position;
	double logWeight;
	QuadraturePoint(double p, double lw){position=p; logWeight=lw;};
};


struct QPoint {
	double x[2],logw;
	unsigned tblIndex;
	unsigned windowIndex;
	double xx[16];
	QPoint(const double x1, const double x2, const double lw, const unsigned index){
		x[0]=x1; x[1]=x2; logw=lw, windowIndex=index;
	};
	bool operator<(const QPoint& other) const{
		return (this->tblIndex < other.tblIndex); 
	};
};


struct QuadratureRule{
	enum type {GaussHermit, GaussLegendre};
	std::vector<QuadraturePoint> rule;

	QuadratureRule(){
		rule.clear(); 
	}
	
	// plain rules
	QuadratureRule(const unsigned order, QuadratureRule::type type=GaussHermit, const double normalFactor=1.0){
		BuildBasicRules(order, type, normalFactor);
	}

	// Gauss-Legendre rules for given mean and range 
	QuadratureRule(const unsigned order, const double mean, const double range, const double normalFactor=1.0){
		BuildBasicRules(order, GaussLegendre, normalFactor);
		for (unsigned int i=0; i<order; i++){
			rule[i].position = (rule[i].position)*range+mean;
			rule[i].logWeight+= log(range);
		}
	}

	//  Gauss-Hermit rules for given background GuassianDistribution
	QuadratureRule(const unsigned order, const GuassianDistribution& g, const double normalFactor=1.0){
		BuildBasicRules(order, GaussHermit, normalFactor); 
		double refMean=g.mean;
		double refI2Var=0.5/g.variance;			
		double fact=sqrt(1.0/refI2Var);
		for (unsigned int i=0; i<order; i++){
			rule[i].position = (rule[i].position)*fact+refMean;
			rule[i].logWeight+= log(fact);
		}
	}

	QuadratureRule& operator+=(const QuadratureRule& beAdded){
		for (unsigned i=0; i<beAdded.rule.size(); i++){
			rule.push_back(beAdded.rule[i]);
		}
		return *this;
	}

private:
	void BuildBasicRules(const unsigned order, QuadratureRule::type type, const double normalFactor){
		rule.clear(); 
		alglib::ae_int_t n=order;
		alglib::ae_int_t info=0;
		alglib::real_1d_array p, w;
		if (n>0) {
			if (type==GaussHermit) alglib::gqgenerategausshermite(n, info, p, w);
			if (type==GaussLegendre) alglib::gqgenerategausslegendre(n, info, p, w);
			for (unsigned i=0; i<order; i++)
				rule.push_back( QuadraturePoint(p(i), log(w(i)*normalFactor) ));
		}
	}


};
typedef std::vector<QuadratureRule> QuadratureRuleSet;

inline void CopyVectorToArray(const NDData& from, alglib::real_1d_array& to){
	to.setlength(from.size());
	for(unsigned i=0; i<from.size(); i++) to(i)=from[i];
}

inline void CopyVectorToArray(const NDDataSet& from, alglib::real_2d_array& to){
	to.setlength(from.size(), from[0].size());
	for(unsigned i=0; i<from.size(); i++) 
		for(unsigned j=0; j<from[0].size(); j++) 
			to(i,j)=from[i][j];
}


#endif // _MathCommon_H
