#include <vector>
#include <algorithm>
#include <limits>
#include <math.h> 
#include <assert.h>

#include <iostream>

#include "alglib/src/interpolation.h"
#include "../include/PhysicalConstants.h"
#include "../include/MathCommon.h"
#include "../include/MathFunctionOnGrid.h"

using namespace alglib;

MathFunctionOnGrid::MathFunctionOnGrid(const OneDDataSet& x, const OneDDataSet& y){

	m_xDataSet.resize(1, x);
	m_yDataSet=y;

}

MathFunctionOnGrid::~MathFunctionOnGrid(void)
{
}

void MathFunctionOnGrid::GetValues(NDDataSet& x, OneDDataSet& y) const{

	x=m_xDataSet;
	y=m_yDataSet;

	return;
}

void MathFunctionOnGrid::GetValues(OneDDataSet& x, OneDDataSet& y) const{

	x=m_xDataSet[0];
	y=m_yDataSet;

	return;
}



MathFunctionOnGrid::MathFunctionOnGrid(const NDData& xmin, const NDData& xmax, const NDInt& numberGrid, void* pSomething,
		void (* GetFunctionValues)(const NDDataSet& xDataSet, OneDDataSet& yDataSet, void* pSomething))
{

	int numberDim=xmax.size();

	int totalNumberGrid=1;
	for (int i=0; i<numberDim; i++){
		totalNumberGrid*=numberGrid[i];
	}

	OneDDataSet pointSet;
	pointSet.resize(totalNumberGrid, 0.0);

	m_xDataSet.resize(numberDim, OneDDataSet(totalNumberGrid, 0.0) );

	PopulateGrid(numberDim-1, xmin, xmax, numberGrid);

	GetFunctionValues(m_xDataSet, m_yDataSet, pSomething);

}

void MathFunctionOnGrid::PopulateGrid(int currentDim, const NDData& xmin, const NDData& xmax, const NDInt& numberGrid){ 

	static int counter=0;

	int numberDim=xmax.size();
	double interval=(xmax[currentDim]-xmin[currentDim])/((numberGrid[currentDim]-1)*1.0);
	if (currentDim>0) {
		for (int i=0; i<numberGrid[currentDim]; i++){
			m_xDataSet[currentDim][counter]=xmin[currentDim]+i*interval;
			PopulateGrid(currentDim-1, xmin, xmax, numberGrid);
			counter++;
		}
	} else { // the lowest dimension
		m_xDataSet[0][counter]=xmin[0];
		for (int i=1; i<numberGrid[0]; i++){
			counter++;		
			for (int j=numberDim-1; j>=currentDim; j--){
				m_xDataSet[j][counter]=m_xDataSet[j][counter-1];
			}
			m_xDataSet[0][counter]=xmin[0]+i*interval;
		}
	}
}


MathFunctionOnGrid MathFunctionOnGrid::Integrate(const MathFunctionOnGrid& derivative){
	
	// support only One-D now

	MathFunctionOnGrid result(derivative);

	int numberGrid=result.m_xDataSet[0].size();
	const double& xmin=result.m_xDataSet[0][0];
	const double& xmax=result.m_xDataSet[0][numberGrid-1];
	double interval=(xmax-xmin)/((numberGrid-1)*1.0);

	result.m_yDataSet[0]=0.0;
	for (int i=1; i<numberGrid; i++){
		result.m_yDataSet[i]+=result.m_yDataSet[i-1];
	}
	for (int i=1; i<numberGrid; i++){
		result.m_yDataSet[i]*=interval;
	}	

	double min=*(
		std::min_element( result.m_yDataSet.begin(), result.m_yDataSet.end()));

	for (int i=0; i<numberGrid; i++){
		result.m_yDataSet[i]-=min;
	}	

	return result;
	
}

double MathFunctionOnGrid::DefiniteIntegral(const MathFunctionOnGrid& function){
	
	// support only One-D now
	return MathFunctionOnGrid::DefiniteIntegral(function.m_xDataSet[0], function.m_yDataSet);

}

double MathFunctionOnGrid::DefiniteIntegral(const OneDDataSet& x, const OneDDataSet& y){
	
	unsigned int numberGrid=x.size();
	const double& xmin=x[0];
	const double& xmax=x[numberGrid-1];
	double interval=(xmax-xmin)/((numberGrid-1)*1.0);

	double result=y[0];

	for (unsigned int i=1; i<numberGrid/2; i++){
		result+=4.0*y[i*2-1];
		result+=2.0*y[i*2];
	}

	result+=y[numberGrid-1];

	return result*interval/3.0;
}


double MathFunctionOnGrid::LogDefiniteIntegralOnExp(const OneDDataSet& x, const OneDDataSet& y){
	
	unsigned numberGrid=x.size();
	assert(y.size()==numberGrid);

	OneDDataSet tempExpY(numberGrid, 0.0);
	double tt=0.0;

	double shift=*(std::max_element(y.begin(), y.end()) );
	for (unsigned i=0; i<numberGrid; i++){
		tt=(y[i]-shift);
		tempExpY[i]= ( (tt>exp_limit) ? exp(tt) : 0.0 );
	}

	return log(MathFunctionOnGrid::DefiniteIntegral(x, tempExpY))+shift;
}


void MathFunctionOnGrid::Gaussian(const double& mean, const double& var, const OneDDataSet& x, OneDDataSet& y, bool normalized){

	unsigned vectorSize = x.size();

	double twoVar=var*2.0;
	double oneOverVar=1.0/var;
	double norm=1.0/sqrt(PhysicalConstants::PI)/sqrt(fabs(twoVar));
	if (!normalized) norm=1.0;

	if (y.size()!=vectorSize) y.resize(vectorSize, 0.0);

	for (unsigned int i=0; i<vectorSize; i++){
		double temp=-(x[i]-mean)*(x[i]-mean)/twoVar;
		if (fabs(temp)<fabs(exp_limit)){
			y[i]=exp(temp)*norm;
		} else {
			y[i]=0.0;
		}
		 
	}

}

void MathFunctionOnGrid::Quadratic(const double& a, const double& x0, const OneDDataSet& x, OneDDataSet& y){
	
	unsigned vectorSize = x.size();
	if (y.size()!=vectorSize) y.resize(vectorSize, 0.0);	
	
	for (unsigned int i=0; i<vectorSize; i++){
		y[i]=(x[i]-x0)*(x[i]-x0)*a;
	}
	
}

void MathFunctionOnGrid::CubicSpline(const OneDDataSet& node, const OneDDataSet& nodeValue, const OneDDataSet& x, OneDDataSet& f, bool periodic){

	static OneDDataSet node_saved;
	static OneDDataSet nodeValue_saved;
	static spline1dinterpolant interpolant;
	unsigned n=node.size();

	if ( (node!=node_saved) || (nodeValue!=nodeValue_saved) ){
	
		real_1d_array xNode;
		real_1d_array fNode;

		CopyVectorToArray(node, xNode);
		CopyVectorToArray(nodeValue, fNode);
		
		// no effect on periodic condition for now
		//if (periodic) {
		//	fNode(n-1)=fNode(0);
		//	spline1dbuildcubic(xNode, fNode, xNode.length(), -1, -1, -1, -1, interpolant);
		//} else {
			spline1dbuildcubic(xNode, fNode, interpolant);
			//spline1dbuildakima(xNode, yNode, interpolant);
		//}

		node_saved=node;
		nodeValue_saved=nodeValue;
	}

	unsigned vectorSize=x.size();
	if (vectorSize>0) {
		f.clear();
		if (f.size()!=vectorSize) f.resize(vectorSize, 0.0);

		for (unsigned i=0; i<vectorSize; i++){
			f[i]=spline1dcalc(interpolant, x[i]);
		}
	} else {
		// put interpolants instead of function values
		real_2d_array tbl;
		tbl.setlength(n-1, 6);

		ae_int_t nn=n;
		spline1dunpack(interpolant, nn, tbl);
		
		f.clear();
		f.resize((n-1)*6, 0.0);

		unsigned counter=0;
		for (unsigned i=0; i<n-1; i++)
			for (unsigned j=0; j<6; j++)
				f[counter++]=tbl(i,j);
	}

	return;

}


void MathFunctionOnGrid::CubicSpline(const NDDataSet& node, const OneDDataSet& nodeValue, const NDDataSet& x, OneDDataSet& f, bool periodic){

	unsigned dim=node.size();
	static NDDataSet node_saved;
	static OneDDataSet nodeValue_saved;
	static spline2dinterpolant interpolant;

	if (dim==1){

		CubicSpline(node[0], nodeValue, x[0], f, periodic);
	
	} else {

		bool sameAsLastTime = ( (node==node_saved) && (nodeValue==nodeValue_saved) );

		if ( !sameAsLastTime ){
			// 2D case only for now
			real_1d_array xNode, yNode, fNode;
			CopyVectorToArray(node[0], xNode);
			CopyVectorToArray(node[1], yNode);
			CopyVectorToArray(nodeValue, fNode);

			unsigned nx=xNode.length();
			unsigned ny=yNode.length();
			spline2dbuildbicubicv(xNode, nx , yNode, ny, fNode, 1, interpolant);
			//spline2dbuildbilinearv(xNode, nx , yNode, ny, fNode, 1, interpolant);
			node_saved=node;
			nodeValue_saved=nodeValue;
		}

		if (x[0].size()>0) {
			real_1d_array ftemp;
			ftemp.setlength(1);
			f.resize(x[0].size(), 0.0);

			for (unsigned i=0; i<x[0].size(); i++){
				spline2dcalcvbuf(interpolant, x[0][i], x[1][i], ftemp);
				f[i]=ftemp(0);
			}

		} else {

			ae_int_t nnX=node[0].size();
			ae_int_t nnY=node[1].size();
			real_2d_array tbl;
			
			ae_int_t d;
			tbl.setlength((nnX-1)*(nnY-1), 20);			
			spline2dunpackv(interpolant, nnX, nnY, d, tbl);
			// At this point, d shoul be one and nnX and nnY should not be changed

			f.clear();
			f.resize(tbl.rows()*tbl.cols(), 0.0);
			double* pF=&(f.at(0));
			unsigned counter=0;
			for (unsigned j=0; j<nnY-1; j++)
				for (unsigned i=0; i<nnX-1; i++)
					for (unsigned k=0; k<20; k++)
						*(pF+counter++)=tbl(i+j*(nnX-1),k);

		}  // if (x.size()>0)

	}  // if (dim==1)....

}

void MathFunctionOnGrid::CubicSpline(const NDDataSet& node, const OneDDataSet& nodeValue, const NDDataSet& x, OneDDataSet& f, NDDataSet& gradient, bool periodic){

	unsigned dim=x.size();
	// limit to 2d for now
	assert(dim<=2);	
	
	unsigned numberDataPoints=x[0].size();

	NDDataSet xtemp(dim, OneDDataSet(0));
	OneDDataSet splineTable;


	// Get the splienTable first
	MathFunctionOnGrid::CubicSpline(node, nodeValue, xtemp, splineTable);

	ae_int_t nnX=node[0].size();
	ae_int_t nnY= (dim>1) ? node[1].size() :2 ;


	f.clear();
	f.resize(numberDataPoints, 0.0);
	gradient.clear();
	gradient.resize(5, OneDDataSet(numberDataPoints, 0.0));

	double* pSF=&(splineTable.at(0));
	double* pF0=&(f.at(0));
	double* pdx0=&(gradient[0].at(0));
	double* pd2x0=&(gradient[1].at(0));
	double* pdy0=&(gradient[2].at(0));
	double* pd2y0=&(gradient[3].at(0));
	double* pdxy0=&(gradient[4].at(0));

	// the x-derivative in gradient[0]
	// the x-2nd-derivative in gradient[1]
	// the y-derivative in gradient[2]
	// the y-2nd-derivative in gradient[3]
	// the xy cross-2nd-derivative in gradient[4]

	unsigned inc=(dim==1) ? 6 : 20;
	unsigned shift=dim*2;
	unsigned nx=4;
	unsigned ny= (dim==1) ? 1 : 4;

	for (unsigned i=0; i<numberDataPoints; i++){
		std::vector<int> index(2, 0);
		std::vector<double> work(2, 0.0);
		for (unsigned idim=0; idim<dim; idim++){
			std::vector<double>::const_iterator lowX=std::lower_bound(node[idim].begin(), node[idim].end(), x[idim][i]);
			int ii=lowX-node[idim].begin()-1;
			if (ii<0) ii=0;
			index[idim]=ii;
			work[idim]=x[idim][i]-node[idim][ii];
		}
		unsigned tblIndex=index[0];
		if (dim>1) tblIndex+=index[1]*(nnX-1);

		unsigned tblPosition=(tblIndex)*inc+shift;
	
		double xx[4][4]={ {1.0, 1.0, 1.0, 1.0},{1.0, 1.0, 1.0, 1.0}, {1.0, 1.0, 1.0, 1.0}, {1.0, 1.0, 1.0, 1.0} };

		for (unsigned iy=1; iy<ny; iy++) xx[0][iy]=xx[0][iy-1]*work[1];
		for (unsigned ix=1; ix<nx; ix++){
			xx[ix][0]=xx[ix-1][0]*work[0];
			for (unsigned iy=1; iy<ny; iy++){
				xx[ix][iy]=xx[ix][iy-1]*work[1];
			}
		}

		double* pF=pF0+i;
		double* pdx=pdx0+i;
		double* pd2x=pd2x0+i;
		double* pdy=pdy0+i;
		double* pd2y=pd2y0+i;
		double* pdxy=pdxy0+i;

		for (unsigned ix=0; ix<nx; ix++) {
			for (unsigned iy=0; iy<ny; iy++){
				double* pp=pSF+iy+ix*ny+tblPosition;
				(*pF)+=xx[ix][iy] * (*pp);
				if (ix>=1) (*pdx)+=xx[ix-1][iy] * (*pp) * ix;
				if (ix>=2) (*pd2x)+=xx[ix-2][iy] * (*pp) *ix*(ix-1);
				if (iy>=1) (*pdy)+=xx[ix][iy-1] * (*pp) * iy;
				if (iy>=2) (*pd2y)+=xx[ix][iy-2] * (*pp) *iy*(iy-1);
				if (ix>=1 && iy>=1) (*pdxy)+=xx[ix-1][iy-1] * (*pp) *ix*iy;				
			} // for (unsigned iy=0; 
		} // for (unsigned ix=0;


	} // for (unsigned i=0; i<numberDataPoints; 

}

void MathFunctionOnGrid::CubicSplineGetZeroGradient(const OneDDataSet& splineTable, const NDData& delta, NDDataSet& zeroGradient, NDDataSet& stationary, const bool maxIncluded){

	unsigned dim=delta.size();

	unsigned inc=(dim==1) ? 6 : 20;
	unsigned shift=dim*2;
	unsigned nx=4;
	unsigned ny= (dim==1) ? 1 : 4;
	unsigned numberSplineNode= splineTable.size()/inc;

	zeroGradient.clear();
	zeroGradient.resize(dim);
	stationary.clear();
	stationary.resize(dim);

	const double* pSF0=&(splineTable.at(0));
	for (unsigned inode=0; inode<numberSplineNode; inode++){

		unsigned tblPosition=inode*inc;
		const double* pSF1=pSF0+tblPosition;
		const double* pSF2=pSF0+tblPosition+shift;
		
		for (unsigned idim=0; idim<dim; idim++){

			// when idim==0, dimIndex[0] is x and dimIndex[1] is y
			// when idim==0, dimIndex[0] is y and dimIndex[1] is x
			unsigned dimIndex[2]={idim, 1-idim};
			double xStart=*(pSF1+dimIndex[0]*2);
			double xEnd=*(pSF1+dimIndex[0]*2+1);

			double yStart=*(pSF1+dimIndex[1]*2);
			double yEnd=*(pSF1+dimIndex[1]*2+1);

			double g_old=1.0e99, x_old=0.0, y_old=0.0;

			if (dim==1){ // 1D case
				double a=0.0, b=0.0, c=0.0, d=0.0, g=0.0;
				double ss[2]={1.0e20, 1.0e20};	
				a= 3*(*(pSF2+3));
				b= 2*(*(pSF2+2));
				c= (*(pSF2+1));
				if (equal(a/b,0.0)){
					ss[0]=-c/b;
				} else {
					d=b*b-4*a*c;
					if (d<0 ) continue;
					ss[0]=(-b+sqrt(d))/(2*a);
					ss[1]=(-b-sqrt(d))/(2*a);
				}

				double xRange=xEnd-xStart;
				//std::cout << "aaa" << a << "\n";
				//std::cout << "bbb" << b << "\n";
				//std::cout << "ccc" << c << "\n";
				//std::cout << "ddd" << d << "\n";
				//std::cout << "xxx" << a/b <<":::::"<< equal(a/b,0.0) << "\n";
				//std::cout << "XXX" << xStart << "\n";
				//std::cout << "XXX" << xEnd << "\n";
				//std::cout << "XXX" << xRange << "\n";
				//std::cout << "XXX" << ss[0] << "\n";
				//std::cout << "XXX" << ss[1] << "\n";
				//std::cout << "=====" << "\n";
				for (int k=0; k<2; k++)
					if (ss[k]>=0 && ss[k]<=xRange) {
						if (maxIncluded || (2*a*ss[k]+b)>=0 )  { 
							stationary[0].push_back(ss[k]+xStart);
						}
					}
			}
			if (dim==2){ // 2D case
				for (double xPos=xStart; xPos<=xEnd; xPos+=delta[0]){

					double x[4]={ 1, xPos-xStart, (xPos-xStart)*(xPos-xStart), (xPos-xStart)*(xPos-xStart)*(xPos-xStart)};
					double a=0.0, b=0.0, c=0.0, g=0.0;
					double ss[2];	

					// solve the gradient=0 point for the second dimension analytically
					for (unsigned ix=0; ix<nx; ix++) {
						unsigned ixy[4]={0,0,0,0};
						if (idim==0) for (unsigned iy=0; iy<ny; iy++) ixy[iy]=iy+ix*ny;
						if (idim==1) for (unsigned iy=0; iy<ny; iy++) ixy[iy]=ix+iy*nx;
						a+= 3*(*(pSF2+ixy[3]))*x[ix];
						b+= 2*(*(pSF2+ixy[2]))*x[ix];
						c+= (*(pSF2+ixy[1]))*x[ix];
					}
					
					if (equal(a/b,0.0)){
						ss[0]=ss[1]=-c/b;
					} else {
						double d=b*b-4*a*c;
						if (d<0 ) continue;
						ss[0]=(-b+sqrt(d))/(2*a);
						ss[1]=(-b-sqrt(d))/(2*a);
					}
					for (unsigned jj=0; jj<2; jj++){
						if (ss[jj]>=0 && ss[jj]<= ( yEnd-yStart) ){

							bool minimum= ( (2*a*ss[jj]+b)>=0  || maxIncluded );
							if ( minimum){
								zeroGradient[dimIndex[0]].push_back(xPos);
								zeroGradient[dimIndex[1]].push_back(ss[jj]+yStart);	

								// calculate the gradient along the first dimension
								double y[4]={1,ss[jj],ss[jj]*ss[jj], ss[jj]*ss[jj]*ss[jj] };
								for (unsigned ix=1; ix<nx; ix++) {
									for (unsigned iy=0; iy<ny; iy++) {
										if (idim==0) g+= *(pSF2+iy+ix*ny)*ix*x[ix-1]*y[iy];
										if (idim==1) g+= *(pSF2+ix+iy*nx)*ix*x[ix-1]*y[iy];
									}
								}
								// if the gradient changes sign, use linear interpolation to 
								// locate the possible stationary point
								bool possibleStationary= (  g*g_old<0 && g_old<1.0e20 );
								if (possibleStationary){
									double xx=x_old-(g_old)/(g-g_old)*(xPos-x_old);
									double yy=y_old-(g_old)/(g-g_old)*(y[1]-y_old);
									stationary[dimIndex[0]].push_back(xx);
									stationary[dimIndex[1]].push_back(yy+yStart);	
								}
								g_old=g;
								x_old=xPos;
								y_old=y[1];
							} 
						}
					} // for (unsigned jj=0;
						
				} // for (double xPos=xStart;
			} //if (dim==2){ 
		} // for (unsigned idim=0


	}

}

void MathFunctionOnGrid::ResampleCubicSpline(const OneDDataSet& inNode, const OneDDataSet& inNodeValue, const OneDDataSet& outNode, OneDDataSet& outNodeValue){

	spline1dinterpolant interpolant;
	unsigned n=inNode.size();
	
	real_1d_array xNode;
	real_1d_array fNode;

	CopyVectorToArray(inNode, xNode);
	CopyVectorToArray(inNodeValue, fNode);
		
	spline1dbuildcubic(xNode, fNode, interpolant);

	unsigned vectorSize=outNode.size();
	if (vectorSize>0) {
		outNodeValue.clear();
		outNodeValue.resize(vectorSize, 0.0);

		for (unsigned i=0; i<vectorSize; i++){
			outNodeValue[i]=spline1dcalc(interpolant, outNode[i]);
		}
	} 

	return;
}

void MathFunctionOnGrid::ResampleCubicSpline(const NDDataSet& inNode, const OneDDataSet& inNodeValue, const NDDataSet& outNode, OneDDataSet& outNodeValue){

	unsigned dim=inNode.size();

	if (dim==1) {

		ResampleCubicSpline(inNode[0], inNodeValue, outNode[0], outNodeValue);

	} else {
		
		spline2dinterpolant interpolant;

		real_1d_array xNode, yNode, fNode;
		CopyVectorToArray(inNode[0], xNode);
		CopyVectorToArray(inNode[1], yNode);
		CopyVectorToArray(inNodeValue, fNode);

		unsigned nx=xNode.length();
		unsigned ny=yNode.length();
		spline2dbuildbicubicv(xNode, nx , yNode, ny, fNode, 1, interpolant);


		real_1d_array ftemp;
		ftemp.setlength(1);
		outNodeValue.resize(outNode[0].size()*outNode[1].size(), 0.0);

		unsigned counter=0;
		for (unsigned j=0; j<outNode[1].size(); j++){
			for (unsigned i=0; i<outNode[0].size(); i++){
				spline2dcalcvbuf(interpolant, outNode[0][i], outNode[1][j], ftemp);
				outNodeValue[counter++]=ftemp(0);
			}
		}

	}

}



void MathFunctionOnGrid::Rational(const OneDDataSet& node, const OneDDataSet& nodeValue, const OneDDataSet& x, OneDDataSet& y, int order, bool periodic){


    real_1d_array xNode;
    real_1d_array yNode;

	unsigned n=node.size();
	xNode.setlength(n);
    yNode.setlength(n);
	
	for (unsigned int i=0; i<n; i++){
		xNode(i)=node[i];
		yNode(i)=nodeValue[i];
	}

	if (periodic){
		yNode(n-1)=yNode(0);
	}

	barycentricinterpolant interpolant;
	barycentricbuildfloaterhormann(xNode, yNode, n, order, interpolant);
	//polynomialbuild(xNode, yNode, interpolant);

	unsigned vectorSize=x.size();
	if (vectorSize>0) {
		y.clear();
		if (y.size()!=vectorSize) y.resize(vectorSize, 0.0);

		for (unsigned i=0; i<vectorSize; i++){
			y[i]=barycentriccalc(interpolant, x[i]);
		}
	} else {
		// put interpolants instead of function values
		// Need to be done: for Rational cases
	}

	return;
}
