#include <stdexcept>

#include <boost/bind.hpp>
#include <boost/function.hpp>

#include <dlib/optimization.h>
#include "alglib/src/optimization.h"

#include "../include/MathCommon.h"
#include "../include/Maximizer.h"

namespace Maximizer_Impl{
	void ff(const alglib::real_1d_array& x, double& func, void* p){

		unsigned n=x.length();
		OneDDataSet arg(n, 0.0);
		for (unsigned i=0; i<n; i++) arg[i]=x(i);
		
		func=-((Maximizer*) p)->FunctionValue(arg);
	};
}

using namespace alglib;
using namespace Maximizer_Impl;

void Maximizer::Maximize(OneDDataSet& parameters){

	if (maxIter <= 0) return;

	converged=false;
	unsigned size=parameters.size();

	real_1d_array x_Alg;
	dlib_ParameterSet x_dLib;

	converged=false;
	switch(lib) {
		case (Maximizer::Alg): 
			CopyVectorToArray(parameters, x_Alg);
			
			try{ 
				Maximize_Alg(x_Alg);
			} catch (std::runtime_error& e){
				for (unsigned i=0; i<size; i++) parameters[i]=x_Alg(i);
				throw e;
			} catch (std::exception& e) {
				for (unsigned i=0; i<size; i++) parameters[i]=x_Alg(i);
				throw e;
			} catch (...){
				for (unsigned i=0; i<size; i++) parameters[i]=x_Alg(i);
				throw std::runtime_error(std::string("Unknown runtime error!"));
			}
			for (unsigned i=0; i<size; i++) parameters[i]=x_Alg(i);
			break;
		
		case (Maximizer::dLib): 
			x_dLib.set_size(size);
			for (unsigned i=0; i<size; i++)	x_dLib(i)=parameters[i];

			try{ 
				Maximize_dLib(x_dLib);
			} catch (std::runtime_error& e){
				for (unsigned i=0; i<size; i++) parameters[i]=x_dLib(i);
				throw e;
			} catch (std::exception& e) {
				for (unsigned i=0; i<size; i++) parameters[i]=x_dLib(i);
				throw e;
			} catch (...){
				for (unsigned i=0; i<size; i++) parameters[i]=x_dLib(i);
				throw std::runtime_error(std::string("Unknown runtime error!"));
			}
			for (unsigned i=0; i<size; i++) parameters[i]=x_dLib(i);
			break;
	}

	converged=true;

}

void Maximizer::Maximize_dLib(dlib_ParameterSet& parameters){

	boost::function<double (const dlib_ParameterSet&)> f
		= boost::bind(&Maximizer::FunctionValue_dLib, this, _1);
	boost::function<dlib_ParameterSet (const dlib_ParameterSet&)> g
		= boost::bind(&Maximizer::Derivative_dLib, this, _1);

	normalFactor = fabs( f(parameters) );

	dlib::find_max_using_approximate_derivatives(
//		dlib::bfgs_search_strategy(),
		dlib::lbfgs_search_strategy(5),
//		dlib::cg_search_strategy(),
		dlib::objective_delta_stop_strategy(tol, maxIter),
		f,
		parameters,
		1e10);
}


void Maximizer::Maximize_Alg(alglib::real_1d_array& x){

    double epsg = 0;
    double diffstep = 1.0e-6;
	double epsx=0;
    minlbfgsstate state;
    minlbfgsreport rep;

	minlbfgscreatef(1, x, diffstep, state);
    minlbfgssetcond(state, epsg, tol, epsx, maxIter);

    minlbfgsoptimize(state, ff, NULL, (void*) this);
    minlbfgsresults(state, x, rep);

}


double Maximizer::FunctionValue_Alg(const alglib::real_1d_array& x){

	unsigned size=x.length();
	OneDDataSet parameters(size, 0.0);
	for (unsigned i=0; i<size; i++) parameters[i]=x(i);
	return FunctionValue(parameters);

}

double Maximizer::FunctionValue_dLib(const dlib_ParameterSet& x){

	unsigned size=x.size();
	OneDDataSet parameters(size, 0.0);
	for (unsigned i=0; i<size; i++) parameters[i]=x(i);
	return FunctionValue(parameters);

}


alglib::real_1d_array Maximizer::Derivative_Alg(const alglib::real_1d_array& x){

	unsigned size=x.length();
	OneDDataSet parameters(size, 0.0);
	for (unsigned i=0; i<size; i++) parameters[i]=x(i);

	OneDDataSet der=Derivative(parameters);

	alglib::real_1d_array der_Alg;
	der_Alg.setlength(size);
	for (unsigned i=0; i<size; i++) der_Alg(i)=der[i];

	return der_Alg;

}

dlib_ParameterSet Maximizer::Derivative_dLib(const dlib_ParameterSet& x){

	unsigned size=x.size();
	OneDDataSet parameters(size, 0.0);
	for (unsigned i=0; i<size; i++) parameters[i]=x(i);

	OneDDataSet der=Derivative(parameters);

	dlib_ParameterSet der_dLib;
	der_dLib.set_size(size);
	for (unsigned i=0; i<size; i++) der_dLib(i)=der[i];

	return der_dLib;

}
