#include <assert.h>
#include <boost/assign/std.hpp>


#include "../include/MathFunctionOnGrid.h"
#include "../include/ModelFunction.h"

using namespace boost::assign;

ModelFunction::ModelFunction(const ModelFunction::functionType type, const int numberParameter)
:m_numberParameter(numberParameter){

	m_initialParameter.clear();
	switch(type){
		case ModelFunction::typeOneD_Order4: 
			{
				m_pFunction=ModelFunction::Order4;
				m_numberParameter=5;
				m_initialParameter += 50.0, -50.0, 1.0, -1.0, 0.5;
			};
			break;
		case ModelFunction::typeOneD_Order6: 
			{
				m_pFunction=ModelFunction::Order6;
				m_numberParameter=7;
				m_initialParameter +=-10, 1, -1, 1,0, -1,0, 1,0, -1,0;
			};
			break;
		case ModelFunction::typeOneD_ExpAndSixOrder: 
			{
				m_pFunction=ModelFunction::ExpAndSixOrder;
				m_numberParameter=8;
				m_initialParameter +=0.0, -10.0, 1.0;
			 };
			break;
		case ModelFunction::typeOneD_ExpCos: 
			{
				m_pFunction=ModelFunction::ExpCos;
				m_numberParameter=5;
				m_initialParameter+= 500,1, -2, 0.2, 0.02;
			 };
			break;
		case ModelFunction::typeOneD_Test: 
			{
				m_pFunction=ModelFunction::Test;
				m_numberParameter=3;
				m_initialParameter += 1,1,1;
			 };
			break;

		case ModelFunction::typeOneD_FT: 
			{
				m_pFunction=ModelFunction::FT;
				m_initialParameter.resize(m_numberParameter);
				m_initialParameter[0]=0.0;
				for (int i=1; i<m_numberParameter/2; i++){
					m_initialParameter[2*i]=0.0;
					m_initialParameter[2*i-1]=1.0;
				}
			 };
			break;
		case ModelFunction::typeOneD_Quadratic: default: 
			{
				m_pFunction=ModelFunction::Quadratic;
				m_numberParameter=2;
				m_initialParameter += 6.0, 0.01;
			 };
			break;

		case ModelFunction::typeOneD_CubicSpline:  
			{
				m_pFunction=ModelFunction::CubicSpline;
				m_initialParameter.clear();
				m_initialParameter.resize(m_numberParameter, 1.0);
				m_initialParameter[0]=0.0;
				for (int i=1; i<m_numberParameter; i++){
					m_initialParameter[i]=sin(6*3.1415926536*i/m_numberParameter);
					//m_initialParameter[i]=(rand() % 10000 )/500.0;
				}
			 };
			break;

		case ModelFunction::typeOneD_CyclicCubicSpline: 
			{
				m_pFunction=ModelFunction::CyclicCubicSpline;
				m_initialParameter.clear();
				m_initialParameter.resize(m_numberParameter, 1.0);
				m_initialParameter[0]=0.0;
				for (int i=1; i<m_numberParameter; i++){
					m_initialParameter[i]=sin(6*3.1415926536*i/m_numberParameter);
				}
			 };
			break;

		case ModelFunction::typeOneD_Rational:  
			{
				m_pFunction=ModelFunction::Rational;
				m_initialParameter.resize(m_numberParameter);
				m_initialParameter[0]=0.0;
				for (int i=1; i<m_numberParameter; i++){
					m_initialParameter[i]=sin(2*3.1415926536*i/m_numberParameter);
				}
			 };
			break;

		 case ModelFunction::typeOneD_CyclicRational: 
			{
				m_pFunction=ModelFunction::CyclicRational;
				m_initialParameter.resize(m_numberParameter);
				m_initialParameter[0]=0.0;
				for (int i=1; i<m_numberParameter; i++){
					m_initialParameter[i]=sin(2*3.1415926536*i/4.0);
				}
			 };
			break;

		case ModelFunction::typeOneD_BEDAM: 
			{
				m_pFunction=ModelFunction::BEDAM;
				m_initialParameter.resize(7);
				m_initialParameter[0]=0.5;// lamda*
				m_initialParameter[1]=0.5;// Pa*
				m_initialParameter[2]=0.5;// Pb*
				m_initialParameter[3]=-5.56;// Ua*
				m_initialParameter[4]=-25.15;// Ub*
				m_initialParameter[5]=2.509;// SigmaA*
				m_initialParameter[6]=2.509;// SigmaB*
			 };
			break;
	}
	m_initialParameter.resize(m_numberParameter, 0.0);

}


void ModelFunction::ExpAndSixOrder(
		const NDDataSet& feature,								
		const std::vector<double>& parameter,
		const NDDataSet& x,
		OneDDataSet& f){
	
	assert(x.size()==1); // Make sure only used in 1-D
	unsigned int size=x[0].size();
	f.clear();
	f.resize(size,parameter[7]);

	for (int k=6; k>=1; k--){
		for (unsigned int i=0; i<size; i++){
			f[i] *= x[0][i];
			f[i] += parameter[k];	
		}
	}
	for (unsigned int i=0; i<size; i++){
			f[i] *= exp(-parameter[0]*fabs(x[0][i]));
	}

	return;
	
}
void ModelFunction::ExpCos(
		const NDDataSet& feature,	
		const std::vector<double>& parameter,
		const NDDataSet& x,
		OneDDataSet& f){ 

	assert(x.size()==1); // Make sure only used in 1-D
	unsigned int size=x[0].size();
	f.clear();
	f.resize(size,parameter[4]);

	for (unsigned int i=0; i<size; i++){
		f[i] += 
		parameter[0]*
		exp(-parameter[1]*x[0][i]+parameter[2])*
		cos(parameter[3]*x[0][i]*3.1415926);
	}

	return;

}

void ModelFunction::Quadratic(
		const NDDataSet& feature,	
		const std::vector<double>& parameter,
		const NDDataSet& x,
		OneDDataSet& f){ 

	assert(x.size()==1); // Make sure only used in 1-D
	unsigned int size=x[0].size();
	f.clear();
	f.resize(size,0.5/parameter[1]);

	for (unsigned int i=0; i<size; i++){
		f[i] *= (x[0][i]-parameter[0])*(x[0][i]-parameter[0]); 
	}
	return;
}

void ModelFunction::Order6(
		const NDDataSet& feature,								
		const std::vector<double>& parameter,
		const NDDataSet& x,
		OneDDataSet& f){ 

	assert(x.size()==1); // Make sure only used in 1-D
	unsigned int size=x[0].size();
	f.clear();
	f.resize(size,parameter[6]);

	for (int k=5; k>=0; k--){
		for (unsigned int i=0; i<size; i++){
			f[i] *= x[0][i];
			f[i] += parameter[k];	
		}
	}

	return;
	
}

void ModelFunction::FT(
		const NDDataSet& feature,								
		const std::vector<double>& parameter,
		const NDDataSet& x,
		OneDDataSet& f){ 

	assert(x.size()==1); // Make sure only used in 1-D	
    static const double factor=3.1415926536/2.0; 

	double lower=feature[0][0];
	double upper=feature[0][1];

	double range=(feature[0][1]-feature[0][0])*factor;
	double middle=(feature[0][1]+feature[0][0])*factor;

	unsigned int size=x[0].size();
    
	OneDDataSet temp;
	temp.resize(size, 0.0);
	for (unsigned int i=0; i<size; i++){
		temp[i]=(x[0][i]-middle)/range;
	}	
	
	f.clear();
	f.resize(size,0.0);

	int n=parameter.size()/2;

	for (int k=0; k<n; k++){
		for (unsigned int i=0; i<size; i++){
			f[i] += parameter[2*k]*sin(k*temp[i]);
			f[i] += parameter[2*k+1]*cos(k*temp[i]);
		}
	}

	return;
	
}

void ModelFunction::Order4(
		const NDDataSet& feature,	
		const std::vector<double>& parameter,
		const NDDataSet& x,
		OneDDataSet& f){ 

	assert(x.size()==1); // Make sure only used in 1-D
	unsigned size=x[0].size();
	f=x[0];

	for (unsigned int i=0; i<size; i++){
		f[i] *= parameter[4];
		f[i] += parameter[3];
	}

	for (int k=2; k>=0; k--){
		for (unsigned int i=0; i<size; i++){
			f[i] *= x[0][i];
			f[i] += parameter[k];
		}
	}

	return;

}


void ModelFunction::Test(
		const NDDataSet& feature,	
		const std::vector<double>& parameter,
		const NDDataSet& x,
		OneDDataSet& f){ 

	assert(x.size()==1); // Make sure only used in 1-D
	unsigned int size=x[0].size();
	f.clear();
	f.resize(size,0.0);

	for (unsigned int i=0; i<size; i++){
		f[i] = parameter[0]/x[0][i] + parameter[1]/pow(x[0][i],6);
	}

	return;

}

void ModelFunction::CubicSpline(
		const NDDataSet& feature,	
		const std::vector<double>& parameter,
		const NDDataSet& x,
		OneDDataSet& f){ 

	unsigned dim=x.size();

	NDDataSet node(dim);
	OneDDataSet value(parameter);

	for (unsigned i=0; i<dim; i++){
		node[i].resize(feature[i].size(), 0.0);
		for (unsigned j=0; j<feature[i].size(); j++){
			node[i][j]=feature[i][j];
		}
	}

	MathFunctionOnGrid::CubicSpline(node, value, x, f, false);

	return;

}

void ModelFunction::CyclicCubicSpline(
		const NDDataSet& feature,	
		const std::vector<double>& parameter,
		const NDDataSet& x,
		OneDDataSet& f){ 

	unsigned dim=x.size();

	NDDataSet node(dim);
	OneDDataSet value(parameter);

	for (unsigned i=0; i<dim; i++){
		node[i].resize(feature[i].size(), 0.0);
		for (unsigned j=0; j<feature[i].size(); j++){
			node[i][j]=feature[i][j];
		}
	}

	MathFunctionOnGrid::CubicSpline(node, value, x, f, true);

	return;

}

void ModelFunction::Rational(
		const NDDataSet& feature,	
		const std::vector<double>& parameter,
		const NDDataSet& x,
		OneDDataSet& f){ 

	assert(x.size()==1); // Make sure only used in 1-D
	unsigned n=feature.size()-2;
	OneDDataSet node, value(parameter);
    node.resize(n, 0.0);

	for (unsigned int i=0; i<n; i++){
		node[i]=feature[0][i+2];
	}

	MathFunctionOnGrid::Rational(node, value, x[0], f, 8);

}

void ModelFunction::CyclicRational(
		const NDDataSet& feature,	
		const std::vector<double>& parameter,
		const NDDataSet& x,
		OneDDataSet& f){ 

	assert(x.size()==1); // Make sure only used in 1-D
	unsigned n=feature.size()-2;
	OneDDataSet node, value(parameter);
    node.resize(n, 0.0);

	for (unsigned int i=0; i<n; i++){
		node[i]=feature[0][i+2];
	}

	MathFunctionOnGrid::Rational(node, value, x[0], f, 8, true);

}

void ModelFunction::BEDAM(
		const NDDataSet& feature,	
		const std::vector<double>& parameter,
		const NDDataSet& x,
		OneDDataSet& f){ 

//parameter[0]=0.5;// lamda*
//parameter[1]=0.5;// Pa*
//parameter[2]=0.5;// Pb*
//parameter[3]=-10.56;// Ua*
//parameter[4]=10.15;// Ub*
//parameter[5]=3.509;// SigmaA*
//parameter[6]=3.509;// SigmaB*

	assert(x.size()==1); // Make sure only used in 1-D
	unsigned size=x[0].size();
	f.resize(size);

	OneDDataSet temp1(size, 0.0), temp2(size, 0.0), temp3(size, 0.0);

	double fa=feature[0][0]*feature[0][0]*parameter[5]*parameter[5];
	double ba=feature[0][0]*parameter[3];
	double fb=feature[0][0]*feature[0][0]*parameter[6]*parameter[6];
	double bb=feature[0][0]*parameter[4];

	double negativeOneOverBeta=-1.0/feature[0][0];

	double t1, ta, tb;
	for (unsigned i=0;i<size; i++){
		t1=parameter[0]-x[0][i];
		ta=parameter[1]*exp(fa*t1*t1+ba*t1);
		tb=parameter[2]*exp(fb*t1*t1+bb*t1);
		f[i]=negativeOneOverBeta*log(ta+tb);
	}
}
