
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "dcdreader.h"

#define CATDCD_MAJOR_VERSION 2
#define CATDCD_MINOR_VERSION 0

#define INIT_NUM_INDICES 4

static int svd_indices(const char *fname, int **indices) {
  int num, max, retval;
  int *ind;
  FILE *fd;

  max = INIT_NUM_INDICES;
  ind = (int *)malloc(INIT_NUM_INDICES*sizeof(int));
  num= 0;
  fd = fopen(fname, "r");
  if (fd == NULL) {
    printf("error opening index file %s\n",fname);
    return 0;
  }
  while ((retval = fscanf(fd, "%d",ind + num)) == 1) {
    num++;
    if (num== max) {
      int *tmp;
      max *= 2;
      tmp = (int *)realloc(ind, max*sizeof(int));
      ind = tmp;
    }
  }
  fclose(fd);
  *indices = ind;
  return num;
}

static void usage(const char *s) {
  printf("       %s -o outputfile [-i indexfile] [-first firstframe]\n"
         "          [-last lastframe] [-stride stride]\n"
         "          inputfile1 inputfile2 ...\n",s);
  printf("       %s -num inputfile1 inputfile2 ...\n", s);
  exit(1);
}

int main(int argc, char *argv[]) {

  DCDhandle h_in, h_out;
  int retval;
  float *x, *y, *z, *xout, *yout, *zout;
  int i, ifile, iset, nset;
  int num_ind, *ind;
  int first, last, stride;
  int just_num = 0;
  char *outfile = NULL;

  printf("CatDCD %d.%d\n", CATDCD_MAJOR_VERSION, CATDCD_MINOR_VERSION);
  if (argc < 3) usage(argv[0]);

  num_ind = 0;
  ind = 0;
  first = last = stride = -1;
  
  argc--;
  argv++;

  while (argc > 1) {
    /* Check to see if an index file has been specified.  */
    if (!strcmp(argv[0],"-i")) {
      printf("Reading indices from %s\n",argv[1]);
      num_ind = svd_indices(argv[1], &ind);
      if (num_ind < 1) {
        printf("Error: no indices read from file %s; exiting...\n",argv[1]);
        free(ind);
        return 1;
      }
      argc -= 2;
      argv += 2;
      continue;
    } else if (!strcmp(argv[0], "-first")) {
      first = atoi(argv[1]);
      argc -= 2;
      argv += 2;
      continue;
    } else if (!strcmp(argv[0], "-last")) {
      last = atoi(argv[1]);
      argc -= 2;
      argv += 2;
      continue;
    } else if (!strcmp(argv[0], "-stride")) {
      stride = atoi(argv[1]);
      argc -= 2;
      argv += 2;
      continue;
    } else if (!strcmp(argv[0], "-o")) {
      outfile = argv[1];
      argc -= 2;
      argv += 2;
      continue;
    } else if (!strcmp(argv[0], "-num")) {
      just_num = 1;
      argc -= 1;
      argv += 1;
      continue;
    }
    /* No flag, so assume that everything else is an input file */
    break;
  }
  
  /* Check to make sure that an output file was specified */
  if (!just_num && !outfile) {
    usage(argv[0]);
  } 

  /* Warn if both -num and -o were specified since no output will be produced*/
  if (just_num && outfile) {
    printf("Warning: -num specified, so no output file will be produced.\n");
  }

  /* 
   * Open the input dcd files so we can read its header.  Need to know how
   * many timesteps there will be, and it's also a good idea to make sure all
   * the files are present and readable before we start filling up the disk...
   */ 

  iset = 0;
  nset = 0;
  for (ifile=0; ifile < argc; ifile++) {
    memset(&h_in, 0, sizeof(DCDhandle));
    retval = dcd_read_open(&h_in, argv[ifile]);
    if (retval) {
      printf("failure opening %s, returned %d.\n",argv[ifile], retval);
      return 1;
    }
    if (just_num) {
      printf("DCD file %s contains %d frames\n", argv[ifile], h_in.NSET); 
    }
    nset += h_in.NSET;
    dcd_read_close(&h_in);
  }

  if (just_num) {
    printf("Total frames: %d\n", nset);
    goto catdcd_normal_exit;
  }

  if (first < 1) first = 1;
  if (last < 0) last = nset;
  if (last < first) {
    printf("Error: last must be greater than or equal to first.\n");
    exit(1);
  }
  if (stride < 0) stride = 1;

  printf("Reading frames %d through %d with stride of %d.\n", first, last, stride);

  nset = 1+(last-first)/stride;

  h_out.N = num_ind ? num_ind : h_in.N;
  h_out.NSET = nset;
  h_out.ISTART = 0;
  h_out.NSAVC = h_in.NSAVC;
  h_out.DELTA = h_in.DELTA;
 
  dcd_write_open(&h_out, outfile);
  printf("Opened file %s for writing %d frames with %d atoms.\n",outfile,
    h_out.NSET, h_out.N);

  x = (float *)malloc(h_in.N * sizeof(float));
  y = (float *)malloc(h_in.N * sizeof(float));
  z = (float *)malloc(h_in.N * sizeof(float));
  xout = (float *)malloc(h_out.N * sizeof(float));
  yout = (float *)malloc(h_out.N * sizeof(float));
  zout = (float *)malloc(h_out.N * sizeof(float));

  for (ifile=0; ifile<argc; ifile++) {
    int numfromfile = 0;
    memset(&h_in, 0, sizeof(DCDhandle));
    retval = dcd_read_open(&h_in, argv[ifile]);
    if (retval) {
      printf("failure opening %s, returned %d.\n",argv[ifile], retval);
      return 1;
    }
    printf("Reading %d frames from file %s...",h_in.NSET, argv[ifile]);
    fflush(stdout);
    if (iset + h_in.NSET < first) {
      printf("wrote 0 frames.\n");
      iset += h_in.NSET;
    } else {
      for (i=0; i<h_in.NSET; i++) {
        dcd_read_step(&h_in, h_in.ISTART + i*h_in.NSAVC, x,y,z);
        iset++;
        if (iset < first) {
          continue;
        }
        if (iset > last) {
          break;
        }
        if ((iset-first) % stride) {
          continue;
        }
        if (num_ind) {
          int j;
          for (j=0; j<num_ind; j++) {
            xout[j] = x[ind[j]];
            yout[j] = y[ind[j]];
            zout[j] = z[ind[j]];
          }
          dcd_write_step(&h_out, xout, yout, zout);
        } else {
          dcd_write_step(&h_out, x, y, z);
        }
        numfromfile++;
      }
      dcd_read_close(&h_in);
      printf("wrote %d frames.\n", numfromfile);
    }
  }
  dcd_write_close(&h_out);

  free(x); 
  free(y); 
  free(z); 
  free(xout); 
  free(yout); 
  free(zout); 
  free(ind);

catdcd_normal_exit:
  printf("CatDCD exited normally.\n");
  return 0;
}

