#include "Utils/include/CommandLineManager.h"
#include "MathTools/include/PhysicalConstants.h"
#include "MathTools/include/MathFunctionOnGrid.h"
#include "MathTools/include/ModelFunction.h"
#include "MathTools/include/StatTools.h"

#include "Common.h"

#include "BEDAMMLEFinder.h"
#include "BEDAMWindow.h"
#include "BEDAMWindowSet.h"

#include "BEDAMSetDecorator.h"

BEDAMSetDecorator::BEDAMSetDecorator(boost::shared_ptr<WindowSet> pSet):SetDecorator(pSet){
	// default
	//m_currentMethod = BEDAMSetDecorator::derivative; 
	m_currentMethod = BEDAMSetDecorator::global; 
}

BEDAMSetDecorator::~BEDAMSetDecorator(void)
{
}

void BEDAMSetDecorator::PerformAnalysis(){

	NDDataSet& x=GetX();
	OneDDataSet& u=GetU();

	NDData xmin, xmax; 
	NDInt numberGrid,totalGrid;
	GetWindowSet()->GetWindowSetRange(xmin, xmax);

	unsigned numberDim=xmin.size();
	numberGrid.resize(numberDim,1000);
	totalGrid.resize(numberDim,1000);

	switch (m_currentMethod){
		case BEDAMSetDecorator::derivative: default:
			  {
					MathFunctionOnGrid derivativeOfPMF(xmin, xmax, totalGrid, (void*) GetWindowSet().get(),
						BEDAMSetDecorator::Derivative);
					MathFunctionOnGrid PMF=MathFunctionOnGrid::Integrate(derivativeOfPMF);
					PMF.GetValues(x[0],u);
			  };
			  break;
		case BEDAMSetDecorator::global: 
			  {
					MathFunctionOnGrid PMF(xmin, xmax, totalGrid, (void*) GetWindowSet().get(),
						BEDAMSetDecorator::GlobalMLEFit);
					PMF.GetValues(x[0],u);
			  };
			  break;
	}

}

void BEDAMSetDecorator::SetupOptions(){

	 GetWindowSet()->SetModelFunctionType(ModelFunction::typeOneD_BEDAM);
}

void BEDAMSetDecorator::Derivative(const NDDataSet& xData, OneDDataSet& U, void* pSet){

	unsigned dataSize=xData[0].size();
	OneDDataSet xTemp=xData[0];
	
	BEDAMWindowSet* pWindowSet= (BEDAMWindowSet*) pSet;
	std::vector<BEDAMWindow*> list=WindowSet::ChangeType<BEDAMWindow>(pWindowSet->GetCurrentWindowList());
	unsigned numberWindows = list.size();

	OneDDataSetCollection rawData(numberWindows);
	OneDDataSet lambda(numberWindows, 0.0);
	OneDDataSet mean(numberWindows, 0.0);
	OneDDataSet variance(numberWindows, 0.0);

	for (unsigned window=0; window<numberWindows; window++){

		BEDAMWindow&  tt= *(list[window]);
		lambda[window]= tt.GetLambda();
		mean[window]=tt.GetObserved()[0].mean;
		variance[window]=sqrt(fabs( tt.GetObserved()[0].variance ));
	}

	MathFunctionOnGrid::CubicSpline(lambda, mean, xData[0], U);

	return;
}


void BEDAMSetDecorator::GlobalMLEFit(const NDDataSet& xData, OneDDataSet& U, void* pSet){

	unsigned dataSize=xData[0].size();
	OneDDataSet xTemp=xData[0];

	OneDDataSet work(dataSize, 0.0);
	
	BEDAMWindowSet* pWindowSet= (BEDAMWindowSet*) pSet;
	std::vector<BEDAMWindow*> list=WindowSet::ChangeType<BEDAMWindow>(pWindowSet->GetCurrentWindowList());
	unsigned numberWindows = list.size();

	OneDDataSetCollection rawData(numberWindows);
	OneDDataSet lambda(numberWindows, 0.0);

	double minMean=1e10, minVariance=5.0;
	double maxMean=-1e10, maxVariance=5.0;

	for (unsigned window=0; window<numberWindows; window++){

		BEDAMWindow&  tt= *(list[window]);
		lambda[window]= tt.GetLambda();
		unsigned dataSize = tt.GetRawData().size();
		rawData[window].resize(dataSize,0.0);
	
		for (unsigned i=0; i<dataSize; i++){
			rawData[window][i]=(tt.GetRawData())[i][0];
		}

		double myMean=tt.GetObserved()[0].mean;
		double myVariance=tt.GetObserved()[0].variance;
		if (myMean<minMean) { minMean=myMean; minVariance=myVariance; };
		if (myMean>maxMean) { maxMean=myMean; maxVariance=myVariance; };

	}

	BEDAMMLEFinder finder(pWindowSet, lambda);

	NDData xmin, xmax;
	pWindowSet->GetDataRange(xmin, xmax);

	double beta=PhysicalConstants::GasConstantInKcalPerMole* (list[0]->GetTemperature()) ;

	NDDataSet feature(1, OneDDataSet(1, 1.0 / beta));
	(finder.GetModelFunction())->SetFeature(feature);

	std::vector<double> parameters((finder.GetModelFunction())->GetInitialParameter());
	parameters[3]=minMean; parameters[4]=maxMean; 
	parameters[5]=sqrt(fabs(minVariance)); parameters[6]=sqrt(fabs(maxVariance)); 

	(finder.GetModelFunction())->SetInitialParameter(parameters);
	finder.FindML(parameters);

	(*(finder.GetFunctionPointer()))(feature, parameters, NDDataSet(1, xTemp), U);

return;
	
}

