#include <stdexcept>
#include "MathTools/include/MiscHelper.h"
#include "MathTools/include/MathFunctionOnGrid.h"
#include "Utils/include/SplineFileHandler.h"
#include "vFEPCommandManager.h"
#include "DiscretePrior.h"

namespace DiscretePrior_Impl{
	void CalculateSplineOnPoints
		(const NDDataSet& currentNodePositions, const OneDDataSet& currentSplineTable,
		const NDDataSet& dataPositions, OneDDataSet& values);
}

void DiscretePrior_Impl::CalculateSplineOnPoints
		(const NDDataSet& currentNodePositions, const OneDDataSet& currentSplineTable,
		const NDDataSet& dataPositions, OneDDataSet& values)
{
	static NDDataSet myNodePositions;
	static NDDataSet myDataPositions;

	static OneDDataSet tblIndex;
	static OneDDataSet xx;
	
	bool newSplineNodes = (myNodePositions != currentNodePositions);
	bool newDataPoints = (myDataPositions != dataPositions);

	if (newSplineNodes) myNodePositions = currentNodePositions;
	if (newDataPoints) myDataPositions = dataPositions;

	unsigned dim = myNodePositions.size();
	unsigned numberXNode = myNodePositions[0].size();
	unsigned numberYNode = (dim>1) ? myNodePositions[1].size() : 1;
	
	unsigned numberData = myDataPositions.size();
	
	if (newSplineNodes || newDataPoints){
		tblIndex.assign(numberData, -1);
		const double delta = 1.0e-4;
		for (unsigned i = 0; i < myDataPositions.size(); i++){
			int myX = -1, myY = 0;
			for (unsigned ix = 0; ix < numberXNode - 1; ix++){
				if ( (myNodePositions[0][ix]-delta) <= myDataPositions[i][0]  && myDataPositions[i][0] < (myNodePositions[0][ix + 1]+delta) ){
					myX = ix;
					break;
				}
			}
			if (dim > 1){
				myY = -1;
				for (unsigned iy = 0; iy < numberYNode - 1; iy++){
					if ((myNodePositions[1][iy]-delta) <= myDataPositions[i][1]  && myDataPositions[i][1] < (myNodePositions[1][iy + 1] +delta) ){
						myY = iy;
						break;
					}
				}
			}
			tblIndex[i] = myX + myY*(numberXNode - 1);
		}
	}

	numberData = myDataPositions.size();
	tblIndex.resize(numberData);

	unsigned nx = 4;
	unsigned ny = (dim>1) ? 4 : 1;
	unsigned tableShift = nx*ny + dim * 2;

	if (newDataPoints){
		xx.assign(numberData*nx*ny, 0.0);
		unsigned index = 0;
		for (unsigned i = 0; i < numberData; i++){
			double x = myDataPositions[i][0];
			x -= currentSplineTable[tblIndex[i] * tableShift];
			double y = (dim>1) ? myDataPositions[i][1] : 1.0;
			if (dim>1) y -= currentSplineTable[tblIndex[i] * tableShift + 2];

			for (unsigned iy = 0; iy < ny; iy++)
				for (unsigned ix = 0; ix < nx; ix++)
					xx[index + ix + iy*nx] = pow(x, ix)*pow(y, iy);

			index += nx*ny;
		}
	}

	values.assign(numberData, 0.0);
	unsigned index = 0;
	for (unsigned i = 0; i < numberData; i++){
		unsigned splineIndex = tblIndex[i] * tableShift + dim * 2;
		
		double tt = 0.0;
		for (unsigned j = 0; j < nx*ny; j++)
			tt += (xx[index + j] * currentSplineTable[splineIndex + j]);

		values[i] = tt;
		index += nx*ny;
	}
}


DiscretePrior::DiscretePrior(const NDDataSet& positions, const OneDDataSet& values, const double weight):
Prior(positions, values, weight)
{
	m_dataPositions = m_nodePositions;
	m_nodePositions.clear();

	Init();
}

DiscretePrior::DiscretePrior(const double weight) : Prior(weight)
{
}

DiscretePrior::~DiscretePrior()
{
}


double DiscretePrior::GetZIntegral(const NDDataSet& currentNodePositions, const OneDDataSet& currentSplineTable) const{

	OneDDataSet values;
	DiscretePrior_Impl::CalculateSplineOnPoints(currentNodePositions, currentSplineTable, m_dataPositions, values);

	unsigned numberData = values.size();
	assert(numberData == m_values.size());

	const double* ff = &(values.at(0));
	const double* f0 = &(m_values.at(0));

	double result = 0.0;
	double tt = 1.0/double(numberData);
	for (unsigned i = 0; i < numberData; i++) result += exp( (f0[i]- ff[i]));

	return log(result)*m_weight;
}

double DiscretePrior::CalculatePriorAverage(const NDDataSet& currentNodePositions, const OneDDataSet& currentSplineTable) const{
	
	OneDDataSet values;
	DiscretePrior_Impl::CalculateSplineOnPoints(currentNodePositions, currentSplineTable, m_dataPositions, values);

	unsigned numberData = values.size();
	assert(numberData == m_values.size());

	const double* ff = &(values.at(0));

	double result = 0.0;
	for (unsigned i = 0; i < numberData; i++) result +=  (*(ff + i));

	return result / double(numberData) *m_weight;

}
void DiscretePrior::Init(){

	unsigned dim=m_dataPositions.size();
	
}