#include <boost/bind.hpp>
#include <boost/function.hpp>

#include "MathTools/include/MathCommon.h"
#include "MathTools/include/MathFunctionOnGrid.h"
#include "Utils/include/Output.h"

#include "vFEPCommandManager.h"
#include "DataWindow.h"
#include "WindowSet.h"
#include "MLEFinder.h"


MLEFinder::MLEFinder(WindowSet* pSet, bool derivative): Maximizer(1000, 1e-5, derivative), m_pSet(pSet){

	std::vector<DataWindow*>& list=m_pSet->GetCurrentWindowList();
	unsigned numberWindows=list.size();

	unsigned dim=m_pSet->GetDimensionality();

	m_rawData.clear();

	for (unsigned window=0; window<numberWindows; window++){

		unsigned dataSize = list[window]->GetRawData().size();
		NDDataSet rawData; rawData.resize(dim, OneDDataSet(dataSize, 0.0));
		for (unsigned idim=0; idim<dim; idim++){
			for (unsigned i=0; i<dataSize; i++){
				rawData[idim][i]=(list[window]->GetRawData())[i][idim];
			}
		}
		//list[window]->CleanRawData();
		m_rawData.push_back(rawData);
	}

	m_pFunction.reset( new ModelFunction(pSet->GetModelFunctionType(), 3*numberWindows+1) ); 

}

void (* MLEFinder::GetFunctionPointer()) 
		(const NDDataSet& feature,
		const std::vector<double>& parameter, 
		const NDDataSet& x, 
		OneDDataSet& y){

	return GetModelFunction()->GetFunctionPointer(); 
}

void MLEFinder::FindML(std::vector<double>& parameters){
	std::vector<double> multipliers(0, 0.0);
	FindML(parameters, multipliers);
}

void MLEFinder::FindML(std::vector<double>& parameters, std::vector<double>& multipliers){
	
	vFEPCommandManager& manager= vFEPCommandManager::Instance();

	unsigned nMul = multipliers.size();
	parameters.insert(parameters.end(), multipliers.begin(), multipliers.end());

	bool useAlgLib =( manager.GetKey("a") || manager.GetKey("A") );
	if ( manager.GetKey("a") || manager.GetKey("A") ) lib=Maximizer::Alg;
	else lib=Maximizer::dLib;

	if ( manager.GetKey("n") )
		manager.GetValue("n", maxIter);

	if ( manager.GetKey("tol") )
		manager.GetValue("tol", tol);

	std::stringstream ss(std::stringstream::in | std::stringstream::out);
	Output& output = Output::Instance();
	

	try {
		Maximize(parameters);
	} catch (std::runtime_error& e ) {
		ss << e.what() << std::endl;
		ss << "The MLE found may be not optimal." << std::endl;
		output.AddValue("error",ss);
	} catch (...) {
		ss << "Unknown problem occurred" << std::endl;
		output.AddValue("error",ss);
	}

	if (nMul>0) parameters.erase(parameters.end() - nMul, parameters.end());

}

const NDData& MLEFinder::GetWeight() const { return m_pSet->GetWindowWeight(); };
const double MLEFinder::GetTemperature() const { return m_pSet->GetTemperature(); };