#include "Utils/include/Output.h"
#include "Utils/include/Periodicity.h"
#include "Utils/include/SplineFileHandler.h"
#include "MathTools/include/PhysicalConstants.h"
#include "MathTools/include/MathFunctionOnGrid.h"
#include "MathTools/include/MiscHelper.h"

#include "vFEPCommandManager.h"
#include "SetDecorator.h"

namespace SetDecorator_Impl {
	const double temp=300.0;
	double scale=PhysicalConstants::GasConstantInKcalPerMole*temp;
}

using namespace SetDecorator_Impl;

SetDecorator::SetDecorator(boost::shared_ptr<WindowSet> pSet)
{
	m_pSet=pSet;

	unsigned dim=pSet->GetDimensionality();
	m_x.resize(dim);
	for (unsigned i=0; i<dim; i++) m_x[i].clear();
	m_u.clear();

	m_feature.resize(dim);
	for (unsigned i=0; i<dim; i++) m_feature[i].clear();
	m_parameters.clear();

	m_mask.clear();

	m_shift=0.0;

	scale=pSet->GetTemperature()*PhysicalConstants::GasConstantInKcalPerMole;

}

SetDecorator::~SetDecorator(void)
{
}

void SetDecorator::OutputResults(){

	AddJacobian();
	OutputFreeEnergyAndDensity();
	OutputPath();
}

void SetDecorator::OutputFreeEnergyAndDensity(){

	unsigned numberDim = GetWindowSet()->GetDimensionality();

	vFEPCommandManager& manager = vFEPCommandManager::Instance();
	Output& output = Output::Instance();

	double xRef = 1.0e99;
	double yRef = 1.0e99;
	if (manager.GetKey("x0")){
		manager.GetValue("x0", xRef);
		if (manager.GetKey("y0")) manager.GetValue("y0", yRef);
	}

	if (numberDim > 1 && yRef > 1e10) xRef = 1.0e99;
	if (xRef > 1e10){
		double min = 1e99;
		for (unsigned i = 0; i < m_u.size(); i++){
			if (m_mask[i] && (min > m_u[i])) min = m_u[i];
		}
		m_shift = min;
	}
	else {
		NDDataSet xx(2, NDData(1));
		xx[0][0] = xRef;
		if (numberDim > 1) xx[1][0] = yRef;
		OneDDataSet ff(1, 0.0);
		NDDataSet& feature = GetFeature();
		OneDDataSet& parameters = GetParameters();
		MathFunctionOnGrid::CubicSpline(feature, parameters, xx, ff);
		m_shift = ff[0];
	}
	for (unsigned i = 0; i < m_u.size(); i++) m_u[i] -= m_shift;

	// Scaling the coordinates if requested
	double s[2] = { 1.0, 1.0 };
	if (manager.GetKey("s1")) manager.GetValue("s1", s[0]);
	if (manager.GetKey("s2")) manager.GetValue("s2", s[1]);

	double efep_x = 0.0;
	if (manager.GetKey("efep")) manager.GetValue("efep", efep_x);

	double normalFactor = 1.0;	
	double minX = 1e99, maxX = 1e-99;
	double minY = 1e99, maxY = 1e-99;

	for (unsigned i = 0; i < m_x[0].size(); i++){
			if (m_mask[i] && (minX > m_x[0][i])) minX = m_x[0][i];
			if (m_mask[i] && (maxX < m_x[0][i])) maxX = m_x[0][i];
	}
	normalFactor *= (maxX - minX)*s[0];
	if (numberDim > 1){
			double minY = 1e99, maxY = 1e-99;
			for (unsigned i = 0; i < m_x[1].size(); i++){
				if (m_mask[i] && (minY > m_x[1][i])) minY = m_x[1][i];
				if (m_mask[i] && (maxY < m_x[1][i])) maxY = m_x[1][i];
		}
			normalFactor *= (maxY - minY)*s[1];
	}

	unsigned dataSize = m_x[0].size();
	bool doProb = (manager.GetKey("prob") || manager.GetKey("dfep") || manager.GetKey("efep"));


	// calculate the probability; find the optimal lambda if requested (eFEP)
	std::vector<double> prob(dataSize, 0.0);
	double xAve_0 = 0.0, xAve=0.0;
	double variance = 0.0;
	double divergence = 0.0; 
	double lambda = 0.0;

	unsigned efep_counter = 0;
	if (doProb){
		double ptot = 1.0;
		MiscHelper::CalculateProbability(m_x, m_u, m_mask, ptot, 0, prob, xAve_0, variance, divergence);

		xAve = xAve_0;

		if (manager.GetKey("efep")){

			lambda = (xAve - efep_x) / variance;
			double leftL = lambda / 2;
			double rightL = lambda * 2;
			double variance_temp = 0.0;

			double leftX = 0.0, rightX = 0.0;
			MiscHelper::CalculateProbability(m_x, m_u, m_mask, ptot, leftL, prob, leftX, variance_temp, divergence);
			MiscHelper::CalculateProbability(m_x, m_u, m_mask, ptot, rightL, prob, rightX, variance_temp, divergence);
			double dL = leftX - efep_x;
			double dR = rightX - efep_x;

			double dM = 0.0;
			do {
				//lambda = leftL + (rightL - leftL) / (fabs(dL) + fabs(dR)) *fabs(dL);
				if (fabs(dL)<fabs(dR)) lambda = leftL + fabs(dL) / variance;
				else lambda = rightL - fabs(dR) / variance;
			
				MiscHelper::CalculateProbability(m_x, m_u, m_mask, ptot, lambda, prob, xAve, variance, divergence);
				dM = xAve - efep_x;
				if (dM*dL > 0) {
					leftL = lambda;
					dL = dM;
				} else {
					rightL = lambda;
					dR = dM;
				}
			} while (fabs(dM / efep_x) > 1e-14 && efep_counter++<100);
		}
	}


	// Output free energy profile
	{
		std::string FEPfileName("");
		if (manager.GetFileName("fep", FEPfileName)) output.SetFileName("FreeEnergyProfile", FEPfileName);

		bool gnu = manager.GetKey("gnu");

		std::stringstream ss(std::stringstream::in | std::stringstream::out);
		unsigned w = 15;
		unsigned lasti = 0;

		if (!equal(lambda, 0)) {
			double minU = 1e99;
			for (unsigned i = 0; i < dataSize; i++){
				if (m_mask[i]) m_u[i] += lambda*m_x[0][i] * s[0];
				if (minU>m_u[i]) minU = m_u[i];
			}
			for (unsigned i = 0; i < dataSize; i++){
				if (m_mask[i]) m_u[i] -= minU;
			}
		}

		for (unsigned i = 0; i < dataSize; i++){
			if (m_mask[i]){
				ss << std::setiosflags(std::ios::fixed) << std::setw(w) << std::setprecision(8);

				if (gnu && numberDim>1 && lasti > 0)
					if (!equal(m_x[0][i], m_x[0][lasti]) && !equal(m_x[1][i], m_x[1][lasti]))	ss << std::endl;

				for (unsigned j = 0; j < numberDim; j++) ss << m_x[j][i] * s[j] << "  ";
				ss << m_u[i]*scale << std::endl;

				lasti = i;
			}
		}
		output.AddValue("FreeEnergyProfile", ss);
	}

	if (manager.GetKey("efep")) {
		std::stringstream ss(std::stringstream::in | std::stringstream::out);
		unsigned w = 15;
		ss << "\teFEP::iteration = " << efep_counter << std::endl;
		ss << "\teFEP::lambda = ";
		ss << std::setiosflags(std::ios::fixed) << std::setw(w) << std::setprecision(13);
		ss << lambda << std::endl;
		ss << "\teFEP::old x Ave. = ";
		ss << std::setiosflags(std::ios::fixed) << std::setw(w) << std::setprecision(13);
		ss << xAve_0 << std::endl;
		ss << "\teFEP::new x Ave. = ";
		ss << std::setiosflags(std::ios::fixed) << std::setw(w) << std::setprecision(13);
		ss << xAve << std::endl;
		output.AddValue("program log", ss);
	}

	// Output the probability
	std::string probFileName("");
	if (manager.GetFileName("prob", probFileName)){
		output.SetFileName("ProbFile", probFileName);

		std::stringstream ss(std::stringstream::in | std::stringstream::out);
		unsigned w = 20;
		//double xAve = 0.0; 

		for (unsigned i = 0; i < dataSize; i++){
			if (m_mask[i]){
				ss << std::setiosflags(std::ios::fixed) << std::setw(w) << std::setprecision(13);
				for (unsigned j = 0; j < numberDim; j++){
					ss << m_x[j][i] * s[j] << "  ";
				}
				ss << std::setiosflags(std::ios::fixed) << std::setw(w) << std::setprecision(13);
				ss << prob[i] << std::endl;
			}
		}
		output.AddValue("ProbFile", ss);
	}

	// Output the forward/backward free energy in dFEP using the given probability
	if (manager.GetKey("dfep")){
		double tt1 = minX, tt2 = 0.0;
		double trialF1 = 0.0, trialF2 = 0.0;
		for (unsigned i = 0; i < dataSize; i++)	{
			if (m_mask[i]){
				tt2 = (m_x[0][i] - tt1)*s[0] / scale;
				trialF1 += exp(-tt2)*prob[i];
				trialF2 += exp(tt2)*prob[i];
			}
		}

		trialF1 = -scale*log(trialF1) + tt1;
		trialF2 = scale*log(trialF2) + tt1;

		std::stringstream ss(std::stringstream::in | std::stringstream::out);
		unsigned w = 18;
		ss << "Trial Free Energy: " << std::endl;
		ss << std::setiosflags(std::ios::fixed) << std::setw(w) << std::setprecision(13);
		ss << "\tdelta F = " << tt1 << std::endl;
		ss << "\tf:Forward: exp(-bE): \t" << trialF1 << std::endl;
		ss << "\tb:Backward: exp(bE): \t" << trialF2 << std::endl;
		output.AddValue("Free Energy Shift", ss);
	}

	if (manager.GetKey("spo")) {
		std::string fileName("");
		if (manager.GetFileName("spo", fileName)) {
			output.SetFileName("spline output", fileName);
			double temp = 300.0;
			if (manager.GetKey("temp")) manager.GetValue("temp", temp);
			SplineFileHandler::WriteToFile(GetFeature(), GetParameters(), temp);
		}
	}
}

void SetDecorator::OutputPath(){

	const Periodicity::PeriodicalConditionSet& c=Periodicity::Instance().GetConditions();
 
	vFEPCommandManager& manager= vFEPCommandManager::Instance();
	Output& output=Output::Instance();

	unsigned dim=GetWindowSet()->GetDimensionality();
	NDDataSet& feature=GetFeature();
	OneDDataSet& parameters=GetParameters();
	NDDataSet& xData=GetX();

	// Scaling the coordinates if requested
	double s[2]={ 1.0, 1.0 };
	if (manager.GetKey("s1")){
		manager.GetValue("s1", s[0]);
	} 
	if (manager.GetKey("s2")){
		manager.GetValue("s2", s[1]);
	} 

	if (manager.GetKey("path") || manager.GetKey("st")){

		OneDDataSet splineTable;
		NDDataSet grad;

		MathFunctionOnGrid::CubicSpline(feature, parameters, NDDataSet(dim, OneDDataSet(0)), splineTable);
		double tt=xData[0][1]-xData[0][0];
		NDData delta(dim, tt); 

		if (manager.GetKey("pd1")) manager.GetValue("pd1", delta[0]);
		if (dim>1)
			if (manager.GetKey("pd2")) manager.GetValue("pd2", delta[1]);

		NDDataSet zeroGradient(0), stationary(0);
		OneDDataSet Utemp(0);

		MathFunctionOnGrid::CubicSplineGetZeroGradient(splineTable, delta, zeroGradient, stationary, manager.GetKey("max")  );

		if (manager.GetKey("path") && dim >1) {
			std::string fileName("");
			if (manager.GetFileName("path", fileName))
				output.SetFileName("path", fileName);	
			std::stringstream ss(std::stringstream::in | std::stringstream::out);

			ss << std::setiosflags(std::ios::fixed);
			MathFunctionOnGrid::CubicSpline(feature, parameters, zeroGradient, Utemp, grad);

			for (unsigned i=0; i< zeroGradient[0].size(); i++){
				bool test1=c[0].isPeriodic, test2=c[0].isPeriodic; 
				test1= (!test1 || (zeroGradient[0][i]>=c[0].min && zeroGradient[0][i]<=c[0].max));
				test2= (!test2 || (zeroGradient[1][i]>=c[1].min && zeroGradient[1][i]<=c[1].max));
				if (test1&& test2){
					double hessian=grad[1][i]*grad[3][i]-grad[4][i]*grad[4][i];
					ss<< std::setw(20) <<  std::setprecision(8) << zeroGradient[0][i]*s[0];
					ss<< std::setw(20) <<  std::setprecision(8) << zeroGradient[1][i]*s[1];
					ss<< std::setw(20) <<  std::setprecision(8) << (Utemp[i]-m_shift)*scale << std::endl;
				}
			}
			output.AddValue("path", ss);
		}
		if (manager.GetKey("st")){
			std::string fileName("");
			if (manager.GetFileName("st", fileName))
				output.SetFileName("stationary", fileName);	
			
			std::stringstream ss(std::stringstream::in | std::stringstream::out);
			ss << std::setiosflags(std::ios::fixed);

			MathFunctionOnGrid::CubicSpline(feature, parameters, stationary, Utemp, grad);
			if (dim==1){
				for (unsigned i=0; i< stationary[0].size(); i++){
					if ( (!c[0].isPeriodic) || (stationary[0][i]>=c[0].min && stationary[0][i]<=c[0].max) ) {
						ss<< std::setw(20) <<  std::setprecision(8) << stationary[0][i]*s[0];
						ss<< std::setw(20) <<  std::setprecision(8) << (Utemp[i]-m_shift)*scale;
						ss<< std::setw(20) <<  std::setprecision(8) << grad[1][i]/s[0]*scale;
						if (grad[1][i]<0) ss<< "\t : Possible local maximum point found" << std::endl;
						if (grad[1][i]>0) ss<< "\t : Possible local minimum point found" << std::endl;
					}
				}
			}
			if (dim==2){
				for (unsigned i=0; i< stationary[0].size(); i++){
						bool test1=c[0].isPeriodic, test2=c[0].isPeriodic; 
						test1= (!test1 || (stationary[0][i]>=c[0].min && stationary[0][i]<=c[0].max));
						test2= (!test2 || (stationary[1][i]>=c[1].min && stationary[1][i]<=c[1].max));
						if (test1 && test2){
							double hessian=grad[1][i]*grad[3][i]-grad[4][i]*grad[4][i];	
							ss<< std::setw(20) <<  std::setprecision(8) << stationary[0][i]*s[0];
							ss<< std::setw(20) <<  std::setprecision(8) << stationary[1][i]*s[1];
							ss<< std::setw(20) <<  std::setprecision(8) << (Utemp[i]-m_shift)*scale;
							ss<< std::setw(20) <<  std::setprecision(8) << hessian/s[0]/s[1]*scale;
							if (hessian<0) ss<< "\t : Possible saddle point found" << std::endl;
							if (hessian>0  && grad[1][i]>0 ) ss<< "\t : Possible local minimum point found" << std::endl;
							if (hessian>0  && grad[1][i]<0 ) ss<< "\t : Possible local maximum point found" << std::endl;
						}
				}
			}
			output.AddValue("stationary", ss);
		}
	}
}

void SetDecorator::AddJacobian(){

	vFEPCommandManager& manager= vFEPCommandManager::Instance();
	Output& output=Output::Instance();

	unsigned dim=GetWindowSet()->GetDimensionality();
	const NDDataSet& feature=GetFeature();
	OneDDataSet& parameters=GetParameters();
	NDDataSet& xData=GetX();

	//Correct the resulting free energy profile
	if (manager.GetKey("jr1")){
		if (*(std::min_element(m_x[0].begin(), m_x[0].end())) >0 ){
			unsigned dataSize=m_x[0].size();
			for (unsigned i=0; i<dataSize; i++)
				m_u[i]+=2.0*log(m_x[0][i]);
		}
	}
	if (manager.GetKey("jr2") && dim>=2){
		if (*(std::min_element(m_x[1].begin(), m_x[1].end())) >0 ){
			unsigned dataSize=m_x[1].size();
			for (unsigned i=0; i<dataSize; i++)
				m_u[i]+=2.0*log(m_x[1][i]);
		}
	}

	//Correct the parameter (assuming cubic-spline type calc)
	if (manager.GetKey("jr1")){
		if (*(std::min_element(feature[0].begin(), feature[0].end())) >0 ){
			unsigned x_count=feature[0].size();
			unsigned y_count=1;
			unsigned counter=0;
			if (dim>=2) y_count=feature[1].size();
			for (unsigned j=0; j<y_count; j++)
				for (unsigned i=0; i<x_count; i++)
					parameters[counter++]+=2.0*log(feature[0][i]);
		} else {
			std::stringstream ss(std::stringstream::in | std::stringstream::out);
			ss<<"\n Jacobian correction requested but negative coordinates of dimesion #1 detected.  No correction will be applied.\n";
			output.AddValue("warning", ss, true);	
		}
	}
	if (manager.GetKey("jr2") && dim>=2){
		if (*(std::min_element(feature[1].begin(), feature[1].end())) >0 ){
			unsigned x_count=feature[0].size();
			unsigned y_count=feature[1].size();
			unsigned counter=0;
			for (unsigned j=0; j<y_count; j++)
				for (unsigned i=0; i<x_count; i++)
					parameters[counter++]+=2.0*log(feature[1][j]);
		} else {
			std::stringstream ss(std::stringstream::in | std::stringstream::out);
			ss<<"\n Jacobian correction requested but negative coordinates of dimesion #2 detected.  No correction will be applied.\n";
			output.AddValue("warning", ss, true);	
		}
	}

}

