#include <stdexcept>
#include "MathTools/include/MiscHelper.h"
#include "MathTools/include/MathFunctionOnGrid.h"
#include "Utils/include/SplineFileHandler.h"
#include "vFEPCommandManager.h"
#include "SplinePrior.h"

SplinePrior::SplinePrior(const NDDataSet& nodePositions, const OneDDataSet& values, const double weight):
Prior(nodePositions, values, weight)
{
	m_backgroundOnly=false;
	Init();
}

SplinePrior::SplinePrior(const double weight) : Prior(weight)
{
	m_backgroundOnly=true;
}

SplinePrior::~SplinePrior()
{
}


double SplinePrior::GetZIntegral(const NDDataSet& currentPositions, const OneDDataSet& currentSplineTable) const{

	if (m_backgroundOnly) {
		return 0.0;
	} else {
		if (!(currentPositions==m_nodePositions && currentSplineTable.size()==m_SplinePriorTable.size()) ) {
			std::string error="**** Prior spline does not match the current settings!";
			std::runtime_error e(error.c_str());
			throw e;
		}
	}

	double result=0.0;
	unsigned dim=currentPositions.size();	

	std::vector<double> logZ(1, 0.0);	
	OneDDataSet tt(currentSplineTable);

	for (unsigned i=0; i<currentSplineTable.size(); i++) tt[i]-=m_SplinePriorTable[i];

	double xAve = 0.0;
	MiscHelper::CalculateLogZ(dim, tt, m_quadraturePointList, 1, logZ, xAve, m_temp/300.0);

	return logZ[0]*m_weight;
}

double SplinePrior::CalculatePriorAverage(const NDDataSet& currentPositions, const OneDDataSet& currentSplineTable) const{
	
	if (!m_backgroundOnly)
		if (!(currentPositions == m_nodePositions && currentSplineTable.size() == m_SplinePriorTable.size())) {
			std::string error="**** Prior spline does not match the current settings!";
			std::runtime_error e(error.c_str());
			throw e;
		}
	
	double result=0.0;
	unsigned dim=currentPositions.size();	
	
	const double* p0=&(currentSplineTable.at(0));
	unsigned nx=4, ny=dim*dim,  shift=2*dim, inc=4*dim*dim+shift;
	
	// Calculate the function value average

	unsigned index_f=shift;
	unsigned counter=0;
	do { 
		const double* p=p0+index_f;
		double dx=0.0, dy=0.0;
		if (dim==1){
			dx=( p[-1]-p[-2]); 
		} else {
			dx=( p[-3]-p[-4]); 
			dy=( p[-1]-p[-2]); 
		}
		for (unsigned i=0; i<nx; i++) 
			for (unsigned j=0; j<ny; j++) result+= (p[j+i*ny])/((i+1.0)*(j+1.0))*pow(dx, i*1.0)*pow(dy,j*1.0);

		index_f+=inc;
		counter++;

	} while ( index_f<currentSplineTable.size() );

	return result/(counter*1.0)*m_weight;

}
void SplinePrior::Init(){

	unsigned dim = m_nodePositions.size();

//  Set up m_SplinePriorTable
	MathFunctionOnGrid::CubicSpline(m_nodePositions, m_values, NDDataSet(dim, OneDDataSet(0)), m_SplinePriorTable);
	
//  Set up m_quadraturePointList
	const unsigned nqXPoint=8;
	const unsigned nqYPoint=8;

	unsigned nx = m_nodePositions[0].size();
	unsigned ny = ((dim == 2) ? m_nodePositions[1].size() : 2);

	m_quadraturePointList.clear();	
	unsigned tblCounter=0;

	for (unsigned j=0; j<ny-1; j++){
		QuadratureRule yRule(QuadratureRule(1, 0.0, 0.0));
		if (dim>1){
			double meanY = (m_nodePositions[1][j + 1] + m_nodePositions[1][j]) / 2.0;
			double rangeY = (m_nodePositions[1][j + 1] - m_nodePositions[1][j]) / 2.0;
			yRule=QuadratureRule(nqYPoint, meanY, rangeY);
		} 

		for (unsigned i=0; i<nx-1; i++){
			double meanX = (m_nodePositions[0][i + 1] + m_nodePositions[0][i]) / 2.0;
			double rangeX = (m_nodePositions[0][i + 1] - m_nodePositions[0][i]) / 2.0;
			QuadratureRule xRule(nqXPoint, meanX, rangeX);

			for (unsigned iqy=0; iqy < yRule.rule.size(); iqy++){
				for (unsigned iqx=0; iqx< xRule.rule.size(); iqx++){
					
					double qx=xRule.rule[iqx].position;
					double qy=((dim>1)? yRule.rule[iqy].position : 0.0);

					double qw=xRule.rule[iqx].logWeight;
					if (dim>1) qw+=yRule.rule[iqy].logWeight;

					QPoint qpoint(qx, qy, qw, 0);

					double temp[2] = { (qx - m_nodePositions[0][i]), 0.0 };
					if (dim>1) temp[1] = (qy - m_nodePositions[1][j]);

					unsigned nx=4,ny=dim*dim;
					double xx[4][4]={ {1.0, 1.0, 1.0, 1.0},{1.0, 1.0, 1.0, 1.0}, {1.0, 1.0, 1.0, 1.0}, {1.0, 1.0, 1.0, 1.0} };

					for (unsigned iy=1; iy<ny; iy++) xx[0][iy]=xx[0][iy-1]*temp[1];
					for (unsigned ix=1; ix<nx; ix++){
						xx[ix][0]=xx[ix-1][0]*temp[0];
						for (unsigned iy=1; iy<ny; iy++){
							xx[ix][iy]=xx[ix][iy-1]*temp[1];
						}
					}

					for (unsigned ix=0; ix<nx; ix++)
						for (unsigned iy=0; iy<ny; iy++)
							qpoint.xx[iy+ix*ny]=xx[ix][iy];

					qpoint.tblIndex=tblCounter;
					m_quadraturePointList.push_back(qpoint);
				}	// iqy		
			} //  iqx

			tblCounter++;
		} // i 
	} // j

	std::sort(m_quadraturePointList.begin(), m_quadraturePointList.end());
	
}