
#include <time.h>
#include <sstream>
#include <stdexcept>

#include "Utils/include/Output.h"
#include "Utils/include/Periodicity.h"
#include "Utils/include/SignalHandler.h"
#include "MathTools/include/MathCommon.h"
#include "MathTools/include/MathFunctionOnGrid.h"
#include "MathTools/include/PhysicalConstants.h"
#include "MathTools/include/MiscHelper.h"
#include "vFEPCommandManager.h"
#include "USWindowSet.h"
#include "SplinePrior.h"
#include "DiscretePrior.h"
#include "USMLEFinder.h"


namespace USMLEFinder_Impl {

	static double logWeightCut=-100;

	static void MonitorProgress(const double sum);
	static void Rule2QuadPoint(const unsigned dim, const unsigned windowIndex, const QuadratureRuleSet& rule, std::vector<QPoint>& quadraturePointList);
	
}
using namespace USMLEFinder_Impl;

USMLEFinder::USMLEFinder(WindowSet* pSet):
			MLEFinder(pSet, true), 
			m_nodePositions( ((USWindowSet*)pSet)->GetNodePositions() ),
			m_splineNodes( ((USWindowSet*)pSet)->GetSplineNodes() ),
			m_ruleSets( ((USWindowSet*)pSet)->GetQuadratureRuleSets() ),
			m_periodicCopyList( ((USWindowSet*)pSet)->GetPeriodicCopyList() )
{
			}


void USMLEFinder::SetPrior(const NDDataSet& positions, const OneDDataSet& values, const Prior::PriorType type, const double weight){
	
	switch (type){
		case Prior::spline:	m_pPrior.reset(new SplinePrior(positions, values, weight)); break;
		case Prior::discrete:	m_pPrior.reset(new DiscretePrior(positions, values, weight)); break;
	}
}

// Backgrpund-only prior
void USMLEFinder::SetPrior(const double weight){
	m_pPrior.reset(new SplinePrior(weight));
}


double USMLEFinder::FunctionValue (const OneDDataSet& parameters) {

	static bool firstTime=true;
	static unsigned counter = 0;
	static bool useBaye=false;
	static bool dataWeight = false;
	static double sps=1.0e-12;

	static double targetXAve = 0.0;
	static unsigned nMul = 0;

	// Set up local storages
	if (firstTime){
		
		firstTime=false;

		vFEPCommandManager& manager= vFEPCommandManager::Instance();

		useBaye=manager.GetKey("baye");
		dataWeight = manager.GetKey("dw");
		
		if (manager.GetKey("sps")) manager.GetValue("sps",sps);
		if (manager.GetKey("lwc")) manager.GetValue("lwc",logWeightCut);		
		if (manager.GetKey("lagrange")){
			nMul = 1;
			manager.GetValue("lagrange", targetXAve);
		}


		BuildQuadraturePointList(dataWeight, logWeightCut);

	}// if (firstTime)

	unsigned dim=GetRawData()[0].size();			
	
	OneDDataSet splineParameters(parameters);

	if (nMul > 0) splineParameters.erase(splineParameters.end() - nMul, splineParameters.end());

	CopyBoundaryParameters(splineParameters);
	// Make sure things are correct
	assert(m_splineNodes.size() == splineParameters.size());

	// Obtain the spline table
	OneDDataSet splineTable;
	(*(GetModelFunction()->GetFunctionPointer()))(GetModelFunction()->GetFeature(), splineParameters, NDDataSet(dim, OneDDataSet(0)), splineTable);

	double xAve;
	//  Calculate the normalization factors (Z)
	unsigned numberWindow = GetRawData().size();

	//  Add to the total contribution with proper weights
	double sum = 0.0;
	double totalWeight = 0.0;
	
	// Usual vFEP part
	{
		//  The normal Z-integral part
		double t1 = 0.0;
		double weightSum = 0.0;

		MiscHelper::CalculateLogZ(dim, splineTable, m_quadraturePointList, (dataWeight ? 1 : numberWindow), m_logZ, xAve, sps);
		for (unsigned window = 0; window < numberWindow; window++) {
			double weight = GetWeight()[window];
			if (dataWeight) t1 += m_logZ[0] * weight;
			else	t1 += m_logZ[window] * weight;
			weightSum += weight;
		}

		//  Calculate the function value summation of all data points through spline-lookup
		double t2 = MiscHelper::CalculateSplineSummation(dim, splineTable, m_splineNodes, useBaye);

		sum += (t1 + t2);
		totalWeight += weightSum;

	}

	// Add the prior contribution
	if (m_pPrior.get()){
		double t1=m_pPrior->GetZIntegral(GetModelFunction()->GetFeature(), splineTable);
		double t2 = m_pPrior->CalculatePriorAverage(GetModelFunction()->GetFeature(), splineTable);
		double weight = m_pPrior->GetWeight();

		totalWeight += weight;
		sum += (t1 + t2)*weight;
	}

	sum /= totalWeight;

	// Constraints ---TBD: not working for now
	if (nMul > 0 ){
		unsigned last = parameters.size()-1;
		double qq = parameters[last];
		double q1 = qq * (xAve - targetXAve);
		sum += q1;

		if (counter % 100 ==0) std::cout << counter << "  " << qq << "  " << xAve << "  " << targetXAve << "  " << sum << std::endl;
	}
		
	counter++;

	MonitorProgress(sum);

	return -sum;

}

OneDDataSet USMLEFinder::Derivative(const OneDDataSet& arg) const{

	return arg;
}

void USMLEFinder::OutputResults(const std::vector<double>& parameters){

	unsigned numberWindow=GetRawData().size();
	unsigned dim=GetRawData()[0].size();

	vFEPCommandManager& manager= vFEPCommandManager::Instance();
	double scale=PhysicalConstants::GasConstantInKcalPerMole*GetTemperature();

	if (manager.GetKey("Z")) {
		Output& output = Output::Instance();	
		
		std::string fileName("");
		if (manager.GetFileName("Z", fileName))
			output.SetFileName("Free Energy Shift", fileName);	

		std::stringstream ss(std::stringstream::in | std::stringstream::out);
		assert(numberWindow==m_logZ.size());
		double tt=0.0;

		if (manager.GetKey("Z0")) tt=m_logZ[0];
		else {
			for (unsigned i=0; i<numberWindow; i++) tt+=m_logZ[i];
			tt/=(numberWindow*1.0);
		}

		for (unsigned i=0; i<numberWindow; i++) m_logZ[i]-=tt;
		ss <<"\nFree Energy Shift (in kT and kcal/mol) :\n"; 
		for (unsigned i=0; i<numberWindow; i++) {
			ss <<"Window# " << std::setiosflags(std::ios::fixed)<< std::setw(6) << i;
			ss <<"   -logZ = ";
			ss << std::setiosflags(std::ios::scientific)<< std::setw(15) <<  std::setprecision(8);
			ss << -m_logZ[i] ;
			ss << std::setiosflags(std::ios::scientific)<< std::setw(15) <<  std::setprecision(8);
			ss << -m_logZ[i]*scale << std::endl;
		}
		output.AddValue("Free Energy Shift",ss);
	}

	return;
	if (dim!=1) return;  

	//TBD: need to clean up below

	const NDDataSet& feature=GetModelFunction()->GetFeature();

	NDDataSet x(1);
	OneDDataSet y;
	//MathFunctionOnGrid U(m_UBiasingOnIntegrationPoint[0]);
	//U.GetValues(x[0],y);
	//  X holds the coordiantes where y is not really used here

	(*(GetModelFunction()->GetFunctionPointer()))(feature, parameters, x, y);
	//  now y holds the trial function vaules

	unsigned numberGrid=x[0].size();
	std::vector<double> m_logZ(numberWindow, 0.0), 
		liSample(numberWindow, 0.0), liModel(numberWindow, 0.0), liGaussian(numberWindow, 0.0), 
		uAtAverage(numberWindow, 0.0), uSample(numberWindow, 0.0), uModel(numberWindow, 0.0),

		xSampleAverage(numberWindow, 0.0), xTrueAverage(numberWindow, 0.0),
		xdSampleAverage(numberWindow, 0.0), xdTrueAverage(numberWindow, 0.0),
		KLDivergence(numberWindow, 0.0);
		

	std::vector<double> temp1(numberGrid,0.0), temp2(numberGrid,0.0), temp3(numberGrid,0.0);

	for (unsigned window=0; window<numberWindow; window++){

		//m_UBiasingOnIntegrationPoint[window].GetValues(x,temp1);
		for (unsigned i=0; i<numberGrid; i++){
			temp1[i]=-temp1[i]-y[i];
			temp2[i]=-y[i]*exp(temp1[i]);   // <U> only
			temp3[i]=temp1[i]*exp(temp1[i]);  // <U+W>
		}

		// Calculte the model ensamble averages

		liModel[window]=-m_logZ[window]+MathFunctionOnGrid::DefiniteIntegral(x[0], temp3)/exp(m_logZ[window]);
		uModel[window]=-MathFunctionOnGrid::DefiniteIntegral(x[0], temp2)/exp(m_logZ[window]);

		for (unsigned i=0; i<numberGrid; i++) temp2[i]=x[0][i]*exp(temp1[i]);
		
		xTrueAverage[window]=MathFunctionOnGrid::DefiniteIntegral(x[0], temp2)/exp(m_logZ[window]);

		double tt=xTrueAverage[window];
		for (unsigned i=0; i<numberGrid; i++) temp2[i]=(x[0][i]-tt)*(x[0][i]-tt)*exp(temp1[i]);
		xdTrueAverage[window]=sqrt(MathFunctionOnGrid::DefiniteIntegral(x[0], temp2)/exp(m_logZ[window]));
	
		// Calculte the ensamble averages based on the sampling points

		OneDDataSet rawData_x=GetRawData()[window][0];
		unsigned sampleSize=rawData_x.size();
		OneDDataSet rawData_y(sampleSize,0.0);
		
		// rawData_y is the trial function value at the sampling data point
		(*(GetModelFunction()->GetFunctionPointer()))(feature, parameters, NDDataSet(1, rawData_x), rawData_y);
		std::vector<double> fact(sampleSize, 0.0);

		tt=0.0;
		for (unsigned i=0; i<sampleSize; i++) tt+=rawData_x[i];
		xSampleAverage[window]=tt/(sampleSize*1.0);

		std::vector<double> x1(1,xSampleAverage[window]), u1(1,0.0);
		(*(GetModelFunction()->GetFunctionPointer()))(feature, parameters, NDDataSet(1,x1) , u1);
		uAtAverage[window]=u1[0];

		double t1=0.0, t2=0.0;

		tt=0.0;
		t1=xSampleAverage[window];
		for (unsigned i=0; i<sampleSize; i++) tt+=(rawData_x[i]-t1)*(rawData_x[i]-t1);
		xdSampleAverage[window]=sqrt(tt/(sampleSize*1.0-1.0));

		MathFunctionOnGrid::Gaussian(xSampleAverage[window], xdSampleAverage[window]*xdSampleAverage[window], rawData_x, fact, true);

		t1=0.0; t2=0.0;
		for (unsigned i=0; i<sampleSize; i++){
			t1+=rawData_y[i]; // <U>
			//t2+=rawData_y[i]+m_UBiasingOnDataPoints[window][i]; // <U+W>
		}
		uSample[window]=t1/(sampleSize*1.0);
		liSample[window]=-t2/(sampleSize*1.0)-m_logZ[window];

		liGaussian[window]=-log(sqrt(2*3.1415926536)*xdSampleAverage[window])-(sampleSize-1.0)/(2.0*sampleSize);

		tt=0.0;
		for (unsigned i=0; i<sampleSize; i++) tt+=rawData_x[i];
		xSampleAverage[window]=tt/(sampleSize*1.0);

		t1=xSampleAverage[window];
		tt=0.0;
		for (unsigned i=0; i<sampleSize; i++) tt+=(rawData_x[i]-t1)*(rawData_x[i]-t1);
		xdSampleAverage[window]=sqrt(tt/(sampleSize*1.0-1.0));

		t1=(xSampleAverage[window]-xTrueAverage[window])*(xSampleAverage[window]-xTrueAverage[window]);
		tt=(xdSampleAverage[window]+xdTrueAverage[window])*(xdSampleAverage[window]-xdTrueAverage[window]);
		KLDivergence[window]=(t1+tt)/( 2*xdTrueAverage[window]*xdTrueAverage[window] );
		KLDivergence[window]+=log( xdTrueAverage[window]/xdSampleAverage[window] );
	}

	std::stringstream ss(std::stringstream::in | std::stringstream::out),
		tt(std::stringstream::in | std::stringstream::out);

	double dd=0;
	for (unsigned window=0; window<numberWindow; window++){
		dd-=m_logZ[window]/double(numberWindow);
	}
	for (unsigned window=0; window<numberWindow; window++){
		unsigned w=12;
		ss << std::setw(3) << window << std::setiosflags(std::ios::fixed)<< std::setprecision(5)  
		<< "  L: " << std::setw(w)<< liSample[window] << std::setw(w)<< liModel[window] << std::setw(w)<< liGaussian[window]
		<< "   -logZ: " << std::setw(w) << -m_logZ[window]
		<< "   U: " << std::setw(w) << uAtAverage[window]<< std::setw(w) << uSample[window] << std::setw(w) << uModel[window]
		<< "   X: " << std::setw(w) << xSampleAverage[window] << std::setw(w) << xTrueAverage[window] 
		<< std::setw(w) << xdSampleAverage[window] << std::setw(w) << xdTrueAverage[window] 
		<< "KL:" << std::setw(w) << KLDivergence[window]
		<< std::endl;
		tt << std::setw(15) << -m_logZ[window]-dd;
	}
	Output& output=Output::Instance();
	output.AddValue("details",ss);
	output.AddValue("FreeEnergyshift",tt);

}


void USMLEFinder::BuildQuadraturePointList(const bool dataWeight, const double& logWeightCut){

	unsigned numberWindow = GetRawData().size();
	unsigned dim = GetRawData()[0].size();

	m_quadraturePointList.clear();
	unsigned counter = 0;

	if (dataWeight){
		QuadratureRuleSet ruleSet;

		DataRangeSet dataRange(dim);
		dataRange[0].min = m_splineNodes[0].position[0];
		if (dim >= 2) dataRange[1].min = m_splineNodes[0].position[1];

		unsigned last = m_splineNodes.size()-1;
		dataRange[0].max = m_splineNodes[last].position[0];
		if (dim >= 2) dataRange[1].max = m_splineNodes[last].position[1];
		
		GuassianDistributionSet biasing(2);
		biasing[0].mean = 0.0; biasing[0].variance = 0.0;
		biasing[1] = biasing[0];

		NDData intSize;
		USWindowSet::BuildUpQuadratureRules(dataRange, biasing, ruleSet, intSize);

		Rule2QuadPoint(dim, 0, ruleSet, m_quadraturePointList);
	}
	else {
		for (unsigned window = 0; window < numberWindow; window++)
			Rule2QuadPoint(dim, window, m_ruleSets[window], m_quadraturePointList);
	}

	unsigned numberQuadPoints=m_quadraturePointList.size();
	unsigned numberNodes=m_splineNodes.size();		

	unsigned nNode[2]={ m_nodePositions[0].size() , (dim>1 ? m_nodePositions[1].size() : 1) };

	assert(numberNodes==nNode[0]*nNode[1]);

	for (unsigned iq=0; iq<numberQuadPoints; iq++){
	
		unsigned myNodeIndex[2]={ nNode[0]-1, nNode[1]-1 };

		for (unsigned idim=0; idim<dim; idim++){
			double tt=m_quadraturePointList[iq].x[idim];
			if (tt<=m_nodePositions[idim][1]){
				myNodeIndex[idim]=1;
			} else if (tt>m_nodePositions[idim][nNode[idim]-1]) {
				myNodeIndex[idim]=nNode[idim]-1;
			} else {
				for (unsigned inode=1; inode <nNode[idim]; inode++){
					if (tt>m_nodePositions[idim][inode]) continue;
					else {
						myNodeIndex[idim]=inode;
						break;
					}
				}
			}
		}

		int myNode=myNodeIndex[0];
		if (dim>1) myNode+=myNodeIndex[1]*nNode[0];

		int tblCounter=myNodeIndex[0]-1;
		if (dim>1) tblCounter+=(myNodeIndex[1]-1)*(nNode[0]-1);

		m_quadraturePointList[iq].tblIndex=tblCounter;

		double reference[2];
		reference[0]=m_splineNodes[myNode].previous[0];
		if (dim>1) reference[1]=m_splineNodes[myNode].previous[1];

		double temp[2]={m_quadraturePointList[iq].x[0]-reference[0], 0.0};
		if (dim>1) temp[1]=m_quadraturePointList[iq].x[1]-reference[1];

		unsigned nx=4,ny=dim*dim;
		double xx[4][4]={ {1.0, 1.0, 1.0, 1.0},{1.0, 1.0, 1.0, 1.0}, {1.0, 1.0, 1.0, 1.0}, {1.0, 1.0, 1.0, 1.0} };

		for (unsigned iy=1; iy<ny; iy++) xx[0][iy]=xx[0][iy-1]*temp[1];
		for (unsigned ix=1; ix<nx; ix++){
			xx[ix][0]=xx[ix-1][0]*temp[0];
			for (unsigned iy=1; iy<ny; iy++){
				xx[ix][iy]=xx[ix][iy-1]*temp[1];
			}
		}

		for (unsigned ix=0; ix<nx; ix++)
			for (unsigned iy=0; iy<ny; iy++)
				m_quadraturePointList[iq].xx[iy+ix*ny]=xx[ix][iy];

	}// for (unsigned iq=0; 

	for (int i=m_quadraturePointList.size()-1; i>=0; i--){
		if (m_quadraturePointList[i].logw<logWeightCut)
			m_quadraturePointList.erase(m_quadraturePointList.begin()+i);
	}

	std::sort(m_quadraturePointList.begin(), m_quadraturePointList.end());

}


void USMLEFinder::CopyBoundaryParameters(std::vector<double>& parameters){

	unsigned numberParameter=parameters.size();
	assert(numberParameter==m_periodicCopyList.size());
	for (unsigned i=0; i<numberParameter; i++){
		if (m_periodicCopyList[i]>0){
			//double tt=(parameters[i]+parameters[m_periodicCopyList[i]])/2.0;
			//parameters[i]=tt;
			//parameters[m_periodicCopyList[i]]=tt;
			parameters[i]=parameters[m_periodicCopyList[i]];
		}
	}
}


void USMLEFinder_Impl::MonitorProgress(const double sum){
	
	static unsigned calledNumber=0;

	calledNumber++;

	std::stringstream ss(std::stringstream::in | std::stringstream::out);
	int dd=int(log10(calledNumber*1.0));
	dd =std::max( dd, 2);
	dd =std::min( dd, 4);
	if (calledNumber % int(pow(10.0,dd)) ==0){
		ss << "# " << calledNumber << " : " ;
		ss << std::setiosflags(std::ios::fixed)<< std::setw(15) <<  std::setprecision(8) << sum << std::endl;
	}
	Output& output = Output::Instance();
	output.AddValue("program log",ss);

	if (calledNumber % 100==0){  // Check user interuption every 100 steps
		SignalHandler& h=SignalHandler::Instance();
		if (h.CheckSignal(2)) {
			std::string error="**** User interuption received.\n";
			std::runtime_error e(error.c_str());
			throw e;
		}
	}

}

void USMLEFinder_Impl::Rule2QuadPoint(
	const unsigned dim, 
	const unsigned windowIndex, 
	const QuadratureRuleSet& rule, 
	std::vector<QPoint>& quadraturePointList){

	unsigned nqXPoint = rule[0].rule.size();
	unsigned nqYPoint = ((dim>1) ? rule[1].rule.size() : 1);

	for (unsigned iqy = 0; iqy < nqYPoint; iqy++){
		for (unsigned iqx = 0; iqx < nqXPoint; iqx++){
			unsigned iq = iqx + iqy*nqXPoint;
			double qx = rule[0].rule[iqx].position;
			double qy = ((dim>1) ? rule[1].rule[iqy].position : 0);

			double qw = rule[0].rule[iqx].logWeight;
			if (dim>1) qw += rule[1].rule[iqy].logWeight;

			QPoint qpoint(qx, qy, qw, windowIndex);
			quadraturePointList.push_back(qpoint);
		}
	} // for (unsigned iqy=0;

}
	
