#include <sstream>
#include <boost/algorithm/string.hpp>

#include "MathTools/include/PhysicalConstants.h"
#include "MathTools/include/MathFunctionOnGrid.h"
#include "MathTools/include/ModelFunction.h"

#include "Utils/include/Periodicity.h"
#include "Utils/include/Output.h"
#include "Utils/include/FunctionFileHandler.h"
#include "Utils/include/SplineFileHandler.h"

#include "Common.h"
#include "vFEPCommandManager.h"
#include "USMLEFinder.h"
#include "Prior.h"

#include "USWindow.h"
#include "USWindowSet.h"
#include "USSetDecorator.h"

namespace USSetDecorator_Impl{

	const unsigned gridFor1Dim=200;
	const unsigned gridFor2Dim=200;

	void ParameterRoughGuess(
		const std::vector<SplineNode>& nodes, 
		const unsigned dim, 
		const unsigned nx, 
		const unsigned ny, 
		OneDDataSet& parameters);

	void ParameterRoughGuess(
		const std::vector<SplineNode>& nodes,
		const unsigned dim,
		const unsigned nx,
		const unsigned ny,
		const NDDataSet& pos,
		const OneDDataSet& values,
		OneDDataSet& parameters);

	void TrimDataPoints(const NDDataSet& splinePos, NDDataSet& pos, OneDDataSet& values);
}

using namespace USSetDecorator_Impl;

USSetDecorator::USSetDecorator(boost::shared_ptr<WindowSet> pSet):SetDecorator(pSet){

	// default
	m_PMFMethod = USSetDecorator::MLE; 
	m_InterporationMethod = USSetDecorator::Cubic; 

}

USSetDecorator::~USSetDecorator(void)
{
}

void USSetDecorator::PerformAnalysis(){

	Output& output=Output::Instance();
	output.AddValue("program log", "Beginning the analysis\n", true);

	vFEPCommandManager& manager= vFEPCommandManager::Instance();

	NDData xmin, xmax; 
	NDInt numberGrid,totalGrid,padNumberGrid;
	GetWindowSet()->GetDataRange(xmin, xmax);
	GetGridRange(xmin, xmax, numberGrid, totalGrid, padNumberGrid);

	unsigned numberDim=GetWindowSet()->GetDimensionality();

	NDDataSet& x=GetX();
	OneDDataSet& u=GetU();
	std::vector<bool>& mask=GetMask();

	x.resize(numberDim, OneDDataSet(0,0.0));

	void (*pmfFunction)(const NDDataSet& xData, OneDDataSet& U, void* pSet);
	switch (m_PMFMethod){
		case USSetDecorator::Derivative: 
			pmfFunction=&USSetDecorator::CalculateCombinedDerivative;
			break;
		case USSetDecorator::Wham: 
			pmfFunction=&USSetDecorator::CalculateWhamPMF;
			break;
		case USSetDecorator::MLE: default:
			pmfFunction=&USSetDecorator::CalculateMLEPMF;
			break;
	}

	MathFunctionOnGrid PMF(xmin, xmax, totalGrid, this,	*pmfFunction);
	PMF.GetValues(x,u);

	for (unsigned i=0; i<numberDim; i++){
		x[i].erase(x[i].end()-padNumberGrid[i], x[i].end());
		x[i].erase(x[i].begin(), x[i].begin()+padNumberGrid[i]);
	}
	u.erase(u.end()-padNumberGrid[0], u.end());
	u.erase(u.begin(), u.begin()+padNumberGrid[0]);

	const Periodicity::PeriodicalConditionSet& c=Periodicity::Instance().GetConditions();
	
	mask.clear();
	unsigned dataSize=x[0].size();
	mask.resize(dataSize, true);

	for (unsigned i=0; i<numberDim; i++){
		if (c[i].isPeriodic){
			for (unsigned j=0; j<x[i].size(); j++)
				mask[j]= (mask[j] && (x[i][j]>=c[i].min) &&  (x[i][j]<=c[i].max) );
		}
	}

	output.AddValue("program log", "Analysis finished\n", true);

}


void USSetDecorator::CalculateCombinedDerivative(const NDDataSet& xData, OneDDataSet& U, void* pSet){

	const bool splineFitting=false;

	USSetDecorator* pDecorator=(USSetDecorator*) pSet;
	USWindowSet* pWindowSet= (USWindowSet*) (pDecorator->GetWindowSet().get());

	std::vector<USWindow*> list=WindowSet::ChangeType<USWindow>(pWindowSet->GetCurrentWindowList());
	unsigned numberWindows=list.size();

	OneDDataSet xTemp;

	if (splineFitting){
		
		xTemp.resize(2*numberWindows-1);
		double start=list[0]->GetObserved()[0].mean;
		double end=list[numberWindows-1]->GetBiasing()[0].mean;
		double delta=(end-start)/((numberWindows-1)*2.0);

		for (unsigned i=0; i<numberWindows*2-2; i++) xTemp[i]=start+delta*i;
		xTemp[numberWindows*2-2]=list[list.size()-1]->GetBiasing()[0].mean;

	} else {
		xTemp=xData[0];
	}

	unsigned dataSize=xTemp.size();
	std::vector<double> working(dataSize, 0.0), weight(dataSize, 0.0), UTemp(dataSize, 0.0);

	for (unsigned window=0; window<numberWindows; window++){

		const USWindow& tt= *(list[window]);
		const GuassianDistribution& 
			observed=tt.GetObserved()[0],
			biasing=tt.GetBiasing()[0],
			bestGuess=tt.GetBestGuess()[0];

		const double& x0_best=bestGuess.mean;
		const double& var_best=bestGuess.variance;
		double oneOverVar_best=1.0/var_best;

		// the following is just a weight function for combining the derivatives
		//		MathFunctionOnGrid::Gaussian(biasing.mean, biasing.variance, xTemp, working);
				MathFunctionOnGrid::Gaussian(observed.mean, observed.variance, xTemp, working);
		//		MathFunctionOnGrid::Gaussian(bestGuess.mean, bestGuess.variance, xTemp, working);

		for (unsigned  i=0; i<dataSize; i++){
			double t1=oneOverVar_best*(xTemp[i]-x0_best)*working[i];
			UTemp[i]+=t1;
			weight[i]+=working[i];
		}
	}

	for (unsigned  i=0; i<dataSize; i++) UTemp[i]/=weight[i];
	
	NDDataSet& feature=pDecorator->GetFeature();
	OneDDataSet& parameters=pDecorator->GetParameters();
	feature.resize(1);

	if (splineFitting) MathFunctionOnGrid::CubicSpline(xTemp, UTemp, xData[0], U);
	else U=UTemp;

	MathFunctionOnGrid derivative(xData, U);	
	MathFunctionOnGrid PMF=MathFunctionOnGrid::Integrate(derivative);

	// note: here xTemp should be changed to be the same as xData.
	PMF.GetValues(xTemp, U);
	feature[0]=xData[0];
	parameters=U;
}

void USSetDecorator::CalculateWhamPMF(const NDDataSet & xData, OneDDataSet& U, void* pSet){

	USSetDecorator* pDecorator=(USSetDecorator*) pSet;
	USWindowSet* pWindowSet= (USWindowSet*) (pDecorator->GetWindowSet().get());

	unsigned dataSize=xData[0].size();
	OneDDataSet xTemp=xData[0];
		
	OneDDataSet	weight(dataSize, 0.0), pU(dataSize, 0.0), work(dataSize, 0.0);
	
	std::vector<USWindow*> list=WindowSet::ChangeType<USWindow>(pWindowSet->GetCurrentWindowList());
	unsigned numberWindows = list.size();

	OneDDataSetCollection pObs(numberWindows, work), pBiasing(numberWindows, work);

	for (unsigned window=0; window<numberWindows; window++){
		const USWindow& tt= *(list[window]);
		const GuassianDistribution& 
			observed=tt.GetObserved()[0],
			biasing=tt.GetBiasing()[0],
			bestGuess=tt.GetBestGuess()[0];

		MathFunctionOnGrid::Gaussian(biasing.mean, biasing.variance, xTemp, pBiasing[window]);
		MathFunctionOnGrid::Gaussian(observed.mean, observed.variance, xTemp, pObs[window]);
	}

	std::vector<double> F(numberWindows, 6.0), FOld(numberWindows, 0.0);
	static double tol=std::numeric_limits<double>::epsilon();
	tol*=10;

	unsigned iteration=0;
	double sd=0.0;
	do {
		for (unsigned i=0; i<dataSize; i++){
			pU[i]=0.0; weight[i]=0.0;
		}
		for (unsigned window=0; window<numberWindows; window++){
			const GuassianDistribution& biasing=list[window]->GetBiasing()[0];
			MathFunctionOnGrid::Gaussian(biasing.mean, biasing.variance, xTemp, work, false);
			for (unsigned i=0; i<dataSize; i++){
					pU[i]+=pObs[window][i];
					weight[i]+=work[i]*exp(F[window]);
			}
		}
		double sum=0;
		for (unsigned i=0; i<dataSize; i++){
			if (weight[i]>tol){
				pU[i]/=weight[i];
				sum+=pU[i];
			}
		}

		for (unsigned window=0; window<numberWindows; window++){
			const GuassianDistribution& biasing=list[window]->GetBiasing()[0];
			MathFunctionOnGrid::Gaussian(biasing.mean, biasing.variance, xTemp, work, false);

			for (unsigned i=0; i<dataSize; i++)	work[i]*=pU[i];
			F[window]=-log(MathFunctionOnGrid::DefiniteIntegral(xTemp, work) );

		}
		sd=0;
		for (unsigned window=0; window<numberWindows; window++){
			sd+=(F[window]-FOld[window])*(F[window]-FOld[window]);
		}

		sd=sqrt(sd/(numberWindows*1.0));
		FOld=F;

	} while (++iteration<2000 && sd >0.0000001);

	U.resize(dataSize, 0.0);
	for (unsigned i=0; i<dataSize; i++){
		U[i]=-log(pU[i]);
	}

	NDDataSet& feature=pDecorator->GetFeature();
	OneDDataSet& parameters=pDecorator->GetParameters();
	feature.resize(1);
	feature[0]=xData[0];
	parameters=U;

	return;

}

void USSetDecorator::CalculateMLEPMF(const NDDataSet& xData, OneDDataSet& U, void* pSet){
	
	USSetDecorator* pDecorator=(USSetDecorator*) pSet;
	USWindowSet* pWindowSet= (USWindowSet*) (pDecorator->GetWindowSet().get());

	unsigned dataSize=xData[0].size();
	
	std::vector<USWindow*> list=WindowSet::ChangeType<USWindow>(pWindowSet->GetCurrentWindowList());
	unsigned numberWindows = list.size();

	USMLEFinder finder(pWindowSet);

	unsigned dim=pWindowSet->GetDimensionality();
	NDDataSet& feature=pDecorator->GetFeature();
	OneDDataSet& parameters=pDecorator->GetParameters();

    // Get the node information from the WindowSet
	const NDDataSet& nodePositions=pWindowSet->GetNodePositions();
	for (unsigned i=0; i<dim; i++)
		for (unsigned j=0; j<nodePositions[i].size(); j++) feature[i].push_back(nodePositions[i][j]);

	//(finder.GetModelFunction())->SetFeature(feature);
	unsigned numberParameters=1;
	for (unsigned i=0; i<dim; i++)
		numberParameters*= (feature[i].size());

	parameters.resize(numberParameters, 0.0);
	
	vFEPCommandManager& manager= vFEPCommandManager::Instance();	



	// Cublic spline prior
	bool getSplineInit=false;
	if (manager.GetKey("spi")){
		std::string fileName="";
		if (manager.GetFileName("spi", fileName)) {
			
			double temp = 300.0;
			if (manager.GetKey("temp")) manager.GetValue("temp", temp);
			bool useInputNodes = manager.GetKey("spk");
			SplineFileHandler::ReadFromFile(fileName, feature, parameters, temp, useInputNodes);
			if (useInputNodes) pWindowSet->ResetSplineNodes(feature);
			getSplineInit = true;

			if (manager.GetKey("spw")){
				double priorWeight = -1.0;
				manager.GetValue("spw", priorWeight);
				if (priorWeight < 0) priorWeight = 1.0;
				finder.SetPrior(feature, parameters, Prior::spline, priorWeight);
			}
		}
	}

	// Discrete-point prior
	if (manager.GetKey("dpi")){
		double priorWeight = -1.0;
		if (manager.GetKey("dpw")) manager.GetValue("dpw", priorWeight);
		if (priorWeight<0) priorWeight = 1.0;

		std::string fileName = "";
		if (manager.GetFileName("dpi", fileName)) {
			NDDataSet pos;
			OneDDataSet value;
			double temp = 300.0;
			if (manager.GetKey("temp")) manager.GetValue("temp", temp);
			FunctionFileHandler::ReadFromFile(fileName, pos, value, true, temp);
			TrimDataPoints(feature, pos, value);
			finder.SetPrior(pos, value, Prior::discrete, priorWeight);

			unsigned nx = feature[0].size();
			unsigned ny = (feature.size()>1) ? feature[1].size() : 0;
			ParameterRoughGuess(pWindowSet->GetSplineNodes(), dim, nx, ny, pos, value, parameters);
			getSplineInit = true;
		}
	}

	if (manager.GetKey("spb"))finder.SetPrior(1.0);
			
	if (!getSplineInit) {
		if (!manager.GetKey("np")){
			unsigned nx=feature[0].size();
			unsigned ny=(feature.size()>1)? feature[1].size() : 0 ; 
			ParameterRoughGuess(pWindowSet->GetSplineNodes(), dim, nx, ny, parameters);
		} else {
			OneDDataSet p((finder.GetModelFunction())->GetInitialParameter());
			p.resize(parameters.size(),0.0);
			parameters=p;
		}
		
	}
	(finder.GetModelFunction())->SetFeature(feature);

	unsigned numberMultipliers = (manager.GetKey("lagrange") ? 1 : 0);
	std::vector<double> multipliers(numberMultipliers, 0.0);
	
	finder.FindML(parameters, multipliers);
	finder.CopyBoundaryParameters(parameters);

	(*(finder.GetFunctionPointer()))(feature, parameters, xData , U);

	finder.OutputResults(parameters);

	return;
}

void USSetDecorator::SetupOptions(){

	vFEPCommandManager& manager= vFEPCommandManager::Instance();
	
// PMF method

	std::string method="MLE";
	if (manager.GetValue("t",method)){
		if (boost::iequals(method,"MLE"))m_PMFMethod = USSetDecorator::MLE;
		if (method=="Derivative" || method=="UI" || method=="ui") m_PMFMethod = USSetDecorator::Derivative;
		if (method=="WHAM" || method=="Wham"|| method=="wham") m_PMFMethod = USSetDecorator::Wham;
	}

// Interpolation option
	std::string interpolation="cubic";
	if (m_PMFMethod == USSetDecorator::MLE){
		if (manager.GetValue("i",interpolation)){
			if (interpolation=="cubic" || interpolation=="spline"|| interpolation=="c"|| interpolation=="s") {
				m_InterporationMethod=USSetDecorator::Cubic;
				if (manager.GetKey("r")){
					GetWindowSet()->SetModelFunctionType(ModelFunction::typeOneD_CyclicCubicSpline);
				} else {
					GetWindowSet()->SetModelFunctionType(ModelFunction::typeOneD_CubicSpline);
				}
			} // if (interpolation=="cubic" ||
			if (interpolation=="rational"|| interpolation=="r" ){
				m_InterporationMethod=USSetDecorator::Rational;
				if (manager.GetKey("r")){
					GetWindowSet()->SetModelFunctionType(ModelFunction::typeOneD_CyclicRational);
				} else {
					GetWindowSet()->SetModelFunctionType(ModelFunction::typeOneD_Rational);
				}
			} // if (interpolation=="rational"||
		} // if (manager.GetValue("i",interpolation))
	} // if (m_Method == USSetDecorator::MLE)

}


void USSetDecorator::GetGridRange(NDData& xmin, NDData& xmax, NDInt& numberGrid, NDInt& totalGrid, NDInt& padNumberGrid){

	vFEPCommandManager& manager= vFEPCommandManager::Instance();
	const Periodicity::PeriodicalConditionSet& c=Periodicity::Instance().GetConditions();
	unsigned numberDim=GetWindowSet()->GetDimensionality();

	unsigned g1= ( (numberDim==1) ? gridFor1Dim : gridFor2Dim );
	unsigned g2=g1;

	if (manager.GetKey("grid")){
		manager.GetValue("grid", g1);
		g2=g1;
	}
	if (manager.GetKey("grid1")){
		manager.GetValue("grid1", g1);
	}
	if (manager.GetKey("grid2")){
		manager.GetValue("grid2", g2);
	}

	unsigned gridForEachDim[2]={g1, g2};
    
	if (!c[0].isPeriodic){
		if (manager.GetKey("x")) manager.GetValue("x",xmin[0]);
		if (manager.GetKey("X")) manager.GetValue("X",xmax[0]);
	}
	if (numberDim>1 && !c[1].isPeriodic){
		if (manager.GetKey("y")) manager.GetValue("y",xmin[1]);
		if (manager.GetKey("Y")) manager.GetValue("Y",xmax[1]);
	}

	numberGrid.resize(numberDim,gridForEachDim[0]);
	totalGrid.resize(numberDim,gridForEachDim[0]);
	if (numberDim==2){
		numberGrid[1]=gridForEachDim[1];
		totalGrid[1]=gridForEachDim[1];
	}
	padNumberGrid.resize(numberDim,0);


	double padFactor=0.2;
	if (m_InterporationMethod==USSetDecorator::Rational || m_InterporationMethod==USSetDecorator::Cubic){
		padFactor=0.0;
	}

	if (manager.GetKey("pad")){
		manager.GetValue("pad", padFactor);
	}
	
	if (m_PMFMethod==USSetDecorator::MLE || m_PMFMethod==USSetDecorator::Wham) {
		for (unsigned i=0; i<numberDim; i++){
			if (c[i].isPeriodic) {
				double spacing=(c[i].max-c[i].min)/(numberGrid[i]-1);
				numberGrid[i]=(xmax[i]-xmin[i])/spacing+1;
				double shift=(c[i].min-xmin[i]);
				double dd=shift-floor(shift/spacing)*spacing;
				xmin[i]+=dd;
				xmax[i]=xmin[i]+spacing*(numberGrid[i]-1);
			}
			double pad=(xmax[i]-xmin[i])*padFactor;
			padNumberGrid[i]=numberGrid[i]*padFactor;
			totalGrid[i]=numberGrid[i]+padNumberGrid[i]*2;
			xmin[i]-=pad; xmax[i]+=pad; 
		}
	}
}

void USSetDecorator_Impl::ParameterRoughGuess(
	const std::vector<SplineNode>& nodes, 
	const unsigned dim, 
	const unsigned nx, 
	const unsigned ny, 
	OneDDataSet& parameters){
	
	// Initial guess based on number of data points around each spline node
	unsigned numberParameters=parameters.size();

	double tot=0, tmin=999999, tmax=0;
	for (unsigned inode=0; inode< nodes.size(); inode++){
		tot+=nodes[inode].x[0][0];
		if (nodes[inode].x[0][0]>1e-5){
			if (nodes[inode].x[0][0]<tmin) tmin=nodes[inode].x[0][0];
			if (nodes[inode].x[0][0]>tmax) tmax=nodes[inode].x[0][0];
		}
	}

	for (unsigned inode=0; inode< nodes.size(); inode++){
		double tt=nodes[inode].x[0][0]+pow(tmin/tmax, 2);
		unsigned ix=nodes[inode].index[0];
		unsigned iy=nodes[inode].index[1];
		unsigned myIndex=ix+nx*iy;
		if (ix==0 || (dim==2 && iy ==0) ) continue;
		parameters[myIndex]=-log(tt/tmax);
	}

	if (dim==1){
		parameters[0]=parameters[1];
		for (unsigned ix=1; ix<nx-1; ix++) parameters[ix]=0.5*(parameters[ix]+parameters[ix+1]);
	}

	if (dim==2){
		for (unsigned iy=0; iy<ny; iy++) parameters[iy*nx]=parameters[iy*nx+1];
		for (unsigned ix=0; ix<nx; ix++) parameters[ix]=parameters[ix+nx];

		parameters[0]=parameters[nx+1];
		for (unsigned iy=1; iy<ny-1; iy++){
			for (unsigned ix=1; ix<nx-1; ix++){
				parameters[iy*nx+ix]=0.25*(
					parameters[iy*nx+ix+1]+parameters[(iy+1)*nx+ix]+
					parameters[(iy+1)*nx+ix+1]+parameters[iy*nx+ix]);
			}
		}
	}
}

void USSetDecorator_Impl::ParameterRoughGuess(
	const std::vector<SplineNode>& nodes,
	const unsigned dim,
	const unsigned nx,
	const unsigned ny,
	const NDDataSet& pos,
	const OneDDataSet& values,
	OneDDataSet& parameters){

	unsigned dataSize = pos.size();
	assert(dataSize == values.size());

	for (unsigned inode = 0; inode< nodes.size(); inode++){
		unsigned ix = nodes[inode].index[0];
		unsigned iy = nodes[inode].index[1];
		unsigned myIndex = ix + nx*iy;
		
		double dd = 0.0, tt = 0.0;

		for (unsigned i = 0; i < dataSize; i++){

			double d1 = 1e-5;
			for (unsigned j = 0; j < dim; j++)
				d1+=(nodes[inode].position[j] - pos[i][j])*(nodes[inode].position[j] - pos[i][j]);

			dd += 1/d1;
			tt += values[i] / d1;
		}
		parameters[myIndex] = tt / dd;
	}
}

void USSetDecorator_Impl::TrimDataPoints(const NDDataSet& splinePos, NDDataSet& pos, OneDDataSet& values){

	unsigned dim = splinePos.size();

	assert(pos.size() == values.size());
	for (unsigned i = 0; i < dim; i++){
		double first = splinePos[i][0];
		double last = splinePos[i][splinePos[i].size()-1];

		for (unsigned j = 0; j < pos.size(); j++){
			if (pos[j][i]<first || pos[j][i]>last){
				pos.erase(pos.begin() + j);
				values.erase(values.begin() + j);
				j--;
			}
		}
	}
}