#include <algorithm>
#include <math.h>
#include <string>
#include <iostream>
#include <ctime>
#include <time.h>
#include <boost/random.hpp>

#include <alglib/src/linalg.h>
#include <alglib/src/specialfunctions.h>
#include <alglib/src/fasttransforms.h>
#include <alglib/src/statistics.h>

#include "Utils/include/FileReader.h"
#include "Utils/include/StringParser.h"
#include "Utils/include/Periodicity.h"
#include "Utils/include/Output.h"

#include "MathTools/include/PhysicalConstants.h"
#include "MathTools/include/StatTools.h"
#include "MathTools/include/PeriodicTools.h"
#include "MathTools/include/MathCommon.h"
#include "MathTools/include/MathFunctionOnGrid.h"

#include "vFEPCommandManager.h"
#include "USWindowSet.h"

#include "USWindow.h"

double USWindow::defaultSpacing[2]={-1.0, -1.0};

USWindow::USWindow(const std::string& fileName, const NDData& biasingEqposition, const NDData& forceConst, const double weight, const int dataFreq):
m_eqPosition(biasingEqposition), m_forceConst(forceConst)
{

	SetDimensionality(biasingEqposition.size());
	SetWindowWeight(weight);
	ExtractBiasingInfo(biasingEqposition, forceConst);

	ReadDataFile(fileName, dataFreq);
	
	DoStat(); 
	BuildUpQuadratureRules();

}

USWindow::USWindow(const NDDataSet& rawData, const NDData& biasingEqposition, const NDData& forceConst, const double weight, bool symmetricPair):
m_eqPosition(biasingEqposition), m_forceConst(forceConst), DataWindow(rawData)
{

	SetDimensionality(biasingEqposition.size());
	SetWindowWeight(weight);
	ExtractBiasingInfo(biasingEqposition, forceConst);
	
	if (symmetricPair){
		NDDataSet& dataSet=GetRawData();
		unsigned datapoints=dataSet.size();
				
		if (datapoints>0){
			unsigned dim=dataSet[0].size();
			
			// Swap the coordinates
			for (unsigned i=0; i<datapoints; i++){
				for (unsigned j=0; j< (dim/2); j++){
					unsigned k=(dim-1)-j;
					double tt=dataSet[i][j];
					dataSet[i][j]=dataSet[i][k];
					dataSet[i][k]=tt;
				}
			}

			// Swap the biasing info.
			for (unsigned j=0; j< (dim/2); j++){
				unsigned k=(dim-1)-j;
				GuassianDistribution tt=m_biasing[j];
				m_biasing[j]=m_biasing[k];
				m_biasing[k]=tt;
			}
		}
	}

	DoStat();  
	BuildUpQuadratureRules();
}


std::vector<USWindow*> USWindow::SplitWindow(USWindow* pParentWindow, unsigned numberSplitPerDimension){

	std::vector<USWindow*> childList(0);
	if (!pParentWindow) return childList;
	if (!pParentWindow->TestData()) return childList; 
	if (numberSplitPerDimension<1) return childList; 

	unsigned numberDimension=pParentWindow->GetDimensionality();
	const DataRangeSet& dataRange=pParentWindow->GetDataRange();
	double weight=pParentWindow->GetWindowWeight();
	
	Output& output = Output::Instance();

	NDData forceConst(pParentWindow->GetForceConst());
	bool noBiasing=true;
	for (unsigned idim=0; idim<numberDimension; idim++)
		noBiasing= noBiasing &&  equal(forceConst[idim],0);
		
	if (!noBiasing){
		std::stringstream ss(std::stringstream::in | std::stringstream::out);
		ss <<"Cannot split a window with a non-zero biasing potential.\n";	
		output.AddValue("warning",ss);
		return childList; 
	}

	std::vector< DataRangeSet > newDataRangeSet(numberSplitPerDimension, DataRangeSet(numberDimension));
	for (unsigned idim=0; idim<numberDimension; idim++){
		double xmin=dataRange[idim].min;
		double xmax=dataRange[idim].max;
		defaultSpacing[idim]=(xmax-xmin)/(numberSplitPerDimension*1.0);
		for (unsigned i=0; i<numberSplitPerDimension; i++){
			newDataRangeSet[i][idim].min=xmin+i*defaultSpacing[idim];
			newDataRangeSet[i][idim].max=xmin+(i+1)*defaultSpacing[idim];
		}
	}

	NDDataSet& dataSet = pParentWindow->GetRawData(); 
	unsigned numberDataPoint=pParentWindow->GetRawData().size();
	unsigned nchild=pow(numberSplitPerDimension*1.00001, numberDimension*1.0);

	std::vector<NDDataSet> childData(nchild, NDDataSet(0));
	for (unsigned i=0; i<numberDataPoint; i++){
		const OneDDataSet& thisPoint=pParentWindow->GetRawData()[i];
		int xIndex=-1, yIndex=0;
		for (unsigned ix=0; ix<numberSplitPerDimension; ix++){
			if (newDataRangeSet[ix][0].min<=thisPoint[0] &&
				newDataRangeSet[ix][0].max>thisPoint[0]){
					xIndex=ix;
					break;
			}
		}
		if (numberDimension>1){
			yIndex=-1;
			for (unsigned iy=0; iy<numberSplitPerDimension; iy++){
				if (newDataRangeSet[iy][1].min<=thisPoint[1] &&
					newDataRangeSet[iy][1].max>thisPoint[1]){
						yIndex=iy;
						break;
				}
			}
		}
		if (xIndex>=0 && yIndex>=0){
			childData[xIndex+yIndex*numberSplitPerDimension].push_back(thisPoint);
		}
	}
	std::vector<NDData> middlePoints(nchild, NDData(0));
	switch (numberDimension) {
		case 1:
			for (unsigned ix=0; ix<numberSplitPerDimension; ix++){
				middlePoints[ix].resize(1);
				middlePoints[ix][0]= ( newDataRangeSet[ix][0].min + newDataRangeSet[ix][0].max )/2.0;
			}
			break;
		case 2:
			for (unsigned iy=0; iy<numberSplitPerDimension; iy++){
				for (unsigned ix=0; ix<numberSplitPerDimension; ix++){
					unsigned index=ix+iy*numberSplitPerDimension;
					middlePoints[index].resize(2);
					middlePoints[index][0]= ( newDataRangeSet[ix][0].min + newDataRangeSet[ix][0].max )/2.0;
					middlePoints[index][1]= ( newDataRangeSet[iy][1].min + newDataRangeSet[iy][1].max )/2.0;
				}
			}
			break;
	}


	std::stringstream ss(std::stringstream::in | std::stringstream::out);
	ss << "Trying to create " << nchild << " windows (the parent window will be deleted) :\n\n";
	output.AddValue("program log",ss);

	for (unsigned i=0; i<nchild; i++){
		if (childData[i].size()>0) {
			double myWeight=weight*childData[i].size()/(numberDataPoint*1.0);
			USWindow* pWindow=new USWindow(childData[i], middlePoints[i], NDData(numberDimension, 0.0), myWeight); 
			if (pWindow){
				if (pWindow->TestData()) {
					pWindow->m_biasing=pParentWindow->m_biasing;
					childList.push_back(pWindow);
				} else {
					delete pWindow;
				}
			}
		}
	}

	// destroy the parent window
	delete pParentWindow;

	return childList;
}

USWindow::~USWindow(void)
{
}

void USWindow::ReadDataFile(const std::string& fileName, const int dataFreq){


	Output& output = Output::Instance();
	vFEPCommandManager& manager = vFEPCommandManager::Instance();

	// Read the data file
	FileReader reader(fileName);
	StringParser parser(reader.GetContains(),"\r\n");

	if (manager.GetKey("v")){
		std::stringstream ss(std::stringstream::in | std::stringstream::out);
		ss << " Finished reading raw data.\n";
		output.AddValue("details", ss);
	}

	std::vector<std::string> entries=parser.GetTokens();
	unsigned numberPoints=entries.size();
	for (int i=numberPoints-1; i>=0 ; i-- )	if (entries[i][0]=='#') entries.erase(entries.begin()+i);
	numberPoints=entries.size();

	unsigned numberDimension=GetDimensionality();

	NDDataSet tempDataSet(0);
	std::vector<double> temp(numberDimension+1, 0.0); // extra dimension for data weights

	double dx[2] ={ -1.0e99, -1.0e99};
	double dX[2] ={ 1.0e99, 1.0e99};


	if (manager.GetKey("dx"))manager.GetValue("dx", dx[0]);
	if (manager.GetKey("dX"))manager.GetValue("dX", dX[0]);
	if (manager.GetKey("dy"))manager.GetValue("dx", dx[1]);
	if (manager.GetKey("dY"))manager.GetValue("dX", dX[1]);

	numberPoints = entries.size();
	if (manager.GetKey("head")){
		unsigned headKeep = 0;
		manager.GetValue("head", headKeep);
		if (headKeep>0 && headKeep<numberPoints)
			entries.erase(entries.begin() + headKeep, entries.end());
	}

	numberPoints = entries.size();
	if (manager.GetKey("tail")){
		unsigned tailKeep=0;
		manager.GetValue("tail", tailKeep);
		if (tailKeep>0 && tailKeep<numberPoints)
			entries.erase(entries.begin(), entries.begin()+numberPoints-tailKeep);
	}
	
	bool useDataWeight = (manager.GetKey("dw") || manager.GetKey("dwe") );
	unsigned requiredEntries = numberDimension + 1;

	if (useDataWeight) requiredEntries++; 

	numberPoints=entries.size();
	tempDataSet.resize(numberPoints, temp);

	unsigned counter = 0;
	for (unsigned i=0; i< numberPoints; i++ ){
	
		const std::string& entry=entries[i];
		StringParser parser(entry," \t");
		
		const std::vector<std::string>& tokens=parser.GetTokens();
		if (tokens.size() >= (requiredEntries)){
			bool cut=false;
			for (unsigned j=0; j< numberDimension; j++) {
				temp[j]=atof(tokens[j+1].c_str());
				//temp[j] = std::stof(tokens[j + 1]); // C++11
				if (temp[j]<dx[j] || temp[j]>dX[j]) cut=true;
			}

			temp[numberDimension] = 1.0e99;
			if (useDataWeight){
				//std::cout << entry.c_str() << "\t" << atof(tokens[numberDimension + 1].c_str()) << std::endl;
				temp[numberDimension] = atof(tokens[numberDimension + 1].c_str());
				//temp[numberDimension] = std::stof(tokens[numberDimension + 1]);  // C++11
			}
			if (!cut) tempDataSet[counter++]=temp;
		}  // if ( tokens.size()>requiredEntries
	} // for (int i=0;...

	tempDataSet.resize(counter);
	entries.clear();

	if (manager.GetKey("v")){
		std::stringstream ss(std::stringstream::in | std::stringstream::out);
		ss << " Finished parsing raw data.\n";
		output.AddValue("details", ss);
	}


	numberPoints=tempDataSet.size();

	if (manager.GetKey("dwe")){
		double beta = 1.0 / PhysicalConstants::GasConstantInKcalPerMole / GetTemperature();
		for (unsigned i = 0; i < numberPoints; i++){
			tempDataSet[i][numberDimension] = exp(tempDataSet[i][numberDimension] * beta-10);
		}
	} else if (manager.GetKey("dw")){
		for (unsigned i = 0; i < numberPoints; i++)	tempDataSet[i][numberDimension] = fabs(tempDataSet[i][numberDimension]);
	}
	
	unsigned window_average = 1;
	if (manager.GetKey("wa")){
	
		manager.GetValue("wa", window_average);
		for (unsigned j = 0; j < numberDimension; j++) {
			double ave = 0.0, ave_0;
			for (unsigned k = 0; k < window_average; k++) ave += tempDataSet[k][j];

			ave_0 = tempDataSet[0][j];
			tempDataSet[0][j] = ave / double(window_average);

			for (unsigned i = 1; i < numberPoints - window_average - 1; i++) {
				ave -= ave_0;
				ave_0 = tempDataSet[i][j];
				ave += tempDataSet[i + window_average - 1][j];
				tempDataSet[i][j] = ave / double(window_average);
			}
		}
		tempDataSet.erase(tempDataSet.begin() + numberPoints - window_average +1 , tempDataSet.end());
	}
	else if (manager.GetKey("wb")){
		manager.GetValue("wb", window_average);
		unsigned order = window_average - 1;
		std::vector<double> weight(0, 0.0);

		double ww = alglib::binomialdistribution(0, order, 0.5);
		if (!equal(ww, 0.0)) weight.push_back(ww);

		for (unsigned i = 1; i<window_average; i++){
			ww = alglib::binomialdistribution(i, order, 0.5) - alglib::binomialdistribution(i - 1, order, 0.5);
			if (!equal(ww, 0.0)) weight.push_back(ww);
		}

		window_average = weight.size();
		std::stringstream ss(std::stringstream::in | std::stringstream::out);
		ss << " Number of binomial points used: " << window_average << ".\n\n";

		output.AddValue("program log", ss);

		for (unsigned j = 0; j < numberDimension; j++) {
			for (unsigned i = 0; i < numberPoints - window_average - 1; i++) {
				double ave = 0.0;
				for (unsigned k = 0; k < window_average; k++) ave += tempDataSet[i + k][j] * weight[k];
				tempDataSet[i][j] = ave;
			}
		}

		tempDataSet.erase(tempDataSet.begin() + numberPoints - window_average + 1, tempDataSet.end());
	}

	if (manager.GetKey("v")){
		std::stringstream ss(std::stringstream::in | std::stringstream::out);
		ss << " Finished processing data.\n";
		output.AddValue("details", ss);
	}

	numberPoints = tempDataSet.size();
	if (manager.GetKey("dump")){
		for (unsigned i = 0; i < numberPoints; i++){
			for (unsigned j = 0; j < numberDimension + 1; j++){
				std::cout << tempDataSet[i][j] << "\t";
			}
			std::cout << std::endl;
		}
	}



	NDDataSet& dataSet=GetRawData();
	dataSet.clear();

	unsigned numberPointToUse=(numberPoints/abs(dataFreq));
	std::vector<int> index(numberPointToUse,0);

	counter=0;
	if (dataFreq<0){
		boost::random::mt19937 generator;
		generator.seed(static_cast<unsigned int>(std::time(0)));
		boost::random::uniform_int_distribution<int> distribution(0,numberPoints-1);
		for (unsigned i=0; i< numberPointToUse; i++){
			counter=distribution(generator);
			index[i]=counter;
		}
	} else {
		for (unsigned i=0; i< numberPointToUse; i++){
			index[i]=counter;
			counter+=abs(dataFreq);
		}
	}

	dataSet.resize(numberPointToUse, temp);
	for (unsigned i = 0; i < numberPointToUse; i++) dataSet[i] = tempDataSet[index[i]];

	{
		std::stringstream ss(std::stringstream::in | std::stringstream::out);
		ss << " Total " << numberPointToUse << " points are read. \n\n";
		output.AddValue("program log", ss);
	}

}

bool USWindow::TestData(){

	NDDataSet& positions = GetRawData(); 
	unsigned numberDataPoint=positions.size();
	unsigned numberDimension=( (numberDataPoint>0) ? numberDimension=positions[0].size(): 0 );

	bool IsWindowOK = (numberDimension>0 && numberDataPoint >0);

	vFEPCommandManager& manager = vFEPCommandManager::Instance();
	if (manager.GetKey("del_bad")){
		unsigned dim = GetDimensionality();
		const DataRangeSet& modifiedDataRange = GetDataRange();
		const NDData& eqPosition = GetEqPosition();
		for (unsigned i = 0; i < dim; i++){
			IsWindowOK &= ((eqPosition[i] > modifiedDataRange[i].min) && (eqPosition[i] < modifiedDataRange[i].max));
		}
	}

	return IsWindowOK;
}


void USWindow::PrintDataInfo(){

	unsigned dim=GetDimensionality();
	const DataRangeSet& modifiedDataRange=GetDataRange();
	const NDData& eqPosition=GetEqPosition();
	const NDData& force=GetForceConst();

	std::stringstream ss(std::stringstream::in | std::stringstream::out);

	ss <<"Positioned at : (";	
	for (unsigned i=0; i<dim; i++) ss << "  " << eqPosition[i];
	ss << ") with " << GetRawData().size() << " data points and weight of ";
	ss << GetWindowWeight() <<".\n";

	for (unsigned i=0; i<dim; i++){
		ss<< "Dimenstion#" << i+1<< " :";
		ss<< " Center position: "<< eqPosition[i];
		ss<< ", Force Const: "<< force[i];		
		ss<< ", Data range: "<< modifiedDataRange[i].min << " to "<<modifiedDataRange[i].max;	
		ss<< std::endl;
                ss << "Number of quadrature rules " << " for dimenstion # " << i+1 << " : " << m_ruleSet[i].rule.size() << std::endl;

		if ( (eqPosition[i]<modifiedDataRange[i].min) || (eqPosition[i]>modifiedDataRange[i].max) ){
			ss << "---The data range may be not consistent with the biasing potential.  Please check.\n";
		}

	}
        ss<< std::endl;
	Output& output = Output::Instance();
	output.AddValue("program log",ss);

}


void USWindow::DoStat(){

	NDDataSet& dataSet = GetRawData(); 
	unsigned numberDataPoint=GetRawData().size();
	unsigned numberDimension=GetDimensionality();

	const Periodicity::PeriodicalConditionSet& c=Periodicity::Instance().GetConditions();
	Output& output = Output::Instance();
	vFEPCommandManager& manager= vFEPCommandManager::Instance();

	// All data points need to be wrapped around the biasing centers before doing any statisics
	unsigned numberPoint=dataSet.size();

	for (unsigned i=0; i< numberDimension; i++){
		if (c[i].isPeriodic){
			for (unsigned j=0; j< numberPoint; j++ ) Wrap(m_biasing[i].mean, c[i].period , dataSet[j][i]);
		}
	}

	DataRangeSet dataRange(numberDimension);

	m_observed.resize(numberDimension);
	m_bestGuess.resize(numberDimension);

	std::vector<double> temp(numberDataPoint,0.0);


	if (manager.GetKey("stat")){
			std::stringstream ss(std::stringstream::in | std::stringstream::out);
			ss <<"\nStatistics for the current Window.\n";
	}

	const double delta=1.0e-3;
	for (unsigned i=0; i<numberDimension; i++){

		for (unsigned j=0; j<numberDataPoint; j++) temp[j]=dataSet[j][i];
	
		//  Do some statistics on the data
		NormalDistribution<double> stat(temp);

		m_observed[i].mean=stat.GetMean();
		m_observed[i].variance=stat.GetVariance();

		dataRange[i].min=stat.GetMin();
		dataRange[i].max=stat.GetMax();

		if ((dataRange[i].max-dataRange[i].min)<delta){
			dataRange[i].max=dataRange[i].min+delta;
		}

		// Get the best guess if assuming gaussian distribution
		double t1=m_biasing[i].variance-m_observed[i].variance;
		double t2=m_biasing[i].variance*m_observed[i].variance;
		double t3=m_biasing[i].variance*m_observed[i].mean-m_observed[i].variance*m_biasing[i].mean;

		m_bestGuess[i].mean=t3/t1;
		m_bestGuess[i].variance=t2/t1;

		if (manager.GetKey("stat")){
			std::stringstream ss(std::stringstream::in | std::stringstream::out);
			unsigned w=15;
			ss << std::setiosflags(std::ios::fixed)<< std::setw(w) <<  std::setprecision(8) ;
			ss <<"Dimension #" << i+1<< ":" << std::endl;
			ss << "\tData Mean\t= " << m_observed[i].mean << std::endl;
			ss << "\tData Variance\t= " << m_observed[i].variance << std::endl;
			ss << "\tData S.D.\t= " << sqrt(m_observed[i].variance) << std::endl;
			if (manager.GetKey("v")){
				
				alglib::real_1d_array data;
				data.setlength(temp.size());

				for (unsigned ii = 0; ii < temp.size(); ii++) data(ii) = temp[ii];

				double mean=0.0, variance=0.0, skewness=0.0, kurtosis=0.0;
				alglib::samplemoments(data, temp.size(), mean, variance, skewness, kurtosis);
				
				ss << "\t Mean #2\t= " << mean << std::endl;
				ss << "\t Variance #2\t= " << variance << std::endl;
				ss << "\t Skewness \t= " << skewness << std::endl;
				ss << "\t Kurtosis\t= " << kurtosis << std::endl;
				ss << std::endl;
				ss << "\tBiasing Mean\t= " << m_biasing[i].mean << std::endl;
				ss << "\tBiasing Var.\t= " << m_biasing[i].variance << std::endl;
				ss << "\tBiasing S.D.\t= " << sqrt(m_biasing[i].variance) << std::endl;
				ss << "\tGuess Mean\t= " << m_bestGuess[i].mean << std::endl;
				ss << "\tGuess Var.\t= " << m_bestGuess[i].variance << std::endl;
				ss << "\tGuess S.D.\t= " << fabs(sqrt(m_bestGuess[i].variance)) << std::endl;
			}
			output.AddValue("Statistics",ss);
		}
	}
	SetDataRange(dataRange);

	double beta=1.0/PhysicalConstants::GasConstantInKcalPerMole/GetTemperature();

	bool directFEP=manager.GetKey("dfep");

	// Calculate the exp(-beta*e) average--could be used in FEP calculations
	if (directFEP){

		if (numberDimension != 1) {
			std::string error = "**** dFEP method is only for 1-D case!";
			std::runtime_error e(error.c_str());
			throw e;
		}
		
		assert(numberDimension == 1);

		double tt = m_observed[0].mean;
		OneDDataSet x_temp(numberDataPoint), t1(numberDataPoint), t2(numberDataPoint);

		for (unsigned j=0; j<numberDataPoint; j++) {
			x_temp[j] = (j*1.0) / (1.0*numberDataPoint);
			t1[j]=-beta*dataSet[j][0];
			t2[j]=beta*dataSet[j][0];
		}

		double s1 = MathFunctionOnGrid::LogDefiniteIntegralOnExp(x_temp, t1)/(-beta);
		double s2 = MathFunctionOnGrid::LogDefiniteIntegralOnExp(x_temp, t2)/(beta);

		double u1 = m_observed[0].mean;
		double u2 = beta*m_observed[0].variance / 2;

		double f1 = 0.0, f2 = 0.0, ff=0.0;

		if (manager.GetKey("ffep")){

			alglib::real_1d_array data;
			alglib::complex_1d_array fdata;

			unsigned nseg = 1; // Bartlett's method when nseq>1;

			data.setlength(numberDataPoint / nseg);
			fdata.setlength(numberDataPoint / nseg);

			for (unsigned iseg = 0; iseg < nseg; iseg++){

				unsigned i1 = iseg*numberDataPoint / nseg;
				unsigned i2 = (iseg + 1)*numberDataPoint / nseg;

				ff = 0;
				for (unsigned i = i1; i < i2; i++) {
					data(i - i1) = dataSet[i][0];
					ff += data(i - i1);
				}
				for (unsigned i = i1; i < i2; i++) data(i - i1) -= (ff / double(numberDataPoint*nseg));

				alglib::fftr1d(data, fdata);

				double fn = double(numberDataPoint / nseg);
				for (unsigned i = 0; i < numberDataPoint / nseg; i++) {
					double c2 = (fdata(i).x*fdata(i).x + fdata(i).y*fdata(i).y);
					c2 = sqrt(c2)*2.0*beta / fn;
					ff = log(alglib::besseli0(c2)) / 2.0 / beta;
					f1 += ff;
					if (i < 32) f2 += ff * 2;
				}
			}
			f1 /= double(nseg);
			f2 /= double(nseg);
		}

		if (manager.GetKey("stat")){
			std::stringstream ss(std::stringstream::in | std::stringstream::out);
			unsigned w=11;
			ss << std::endl;
			ss << std::setiosflags(std::ios::fixed)<< std::setw(w) <<  std::setprecision(5);
			ss << " Forward FEP with exp(-beta*E): " << std::endl; 
				ss << "\tf:Direct   = \t" << s1 << std::endl; 
				ss << "\tf:Variance = \t" << u1 - u2 << "\t(" << u1 << "-" << u2 << ")" << std::endl;
				ss << "\tf:fFEP     = \t" << u1 - f1 << "\t(" << u1 << "-" << f1 << ")" << std::endl;
				ss << "\tf:fFEP_cut = \t" << u1 - f2 << "\t(" << u1 << "-" << f2 << ")" << std::endl;
			ss << " Backward FEP with exp(beta*E): " << std::endl; 
				ss << "\tb:Direct   = \t" << s2 << std::endl;
				ss << "\tb:Variance = \t"<<  u1 + u2 << "\t(" << u1 << "+" << u2 << ")" << std::endl;
				ss << "\tb:fFEP     = \t" << u1 + f1 << "\t(" << u1 << "+" << f1 << ")" << std::endl;
				ss << "\tb:fFEP_cut = \t" << u1 + f2 << "\t(" << u1 << "+" << f2 << ")" << std::endl;
			ss << std::endl;
			output.AddValue("Statistics",ss);
		}

	}
}


void USWindow::BuildUpQuadratureRules(){

	NDData intSize;
	USWindowSet::BuildUpQuadratureRules(GetDataRange(), m_biasing, m_ruleSet, intSize);
	SetIntegrationSize(intSize);
}


void USWindow::ExtractBiasingInfo(const NDData& biasingEqposition, const NDData& forceConst){

	// Find the biasing potential centers and constants
	unsigned numberDimension=GetDimensionality();
	const Periodicity::PeriodicalConditionSet& c=Periodicity::Instance().GetConditions();

	double kFactor=1.0;

	vFEPCommandManager& manager= vFEPCommandManager::Instance();
	if (manager.GetKey("k2")) kFactor*=2.0;


	m_biasing.resize(numberDimension);
	for (unsigned i=0; i<numberDimension; i++){

		double forceConstInRT = 
			forceConst[i]/(PhysicalConstants::GasConstantInKcalPerMole*GetTemperature())*kFactor;

		if (!equal(forceConstInRT, 0.0)) m_biasing[i].variance=1.0/forceConstInRT;
		else m_biasing[i].variance=mini_double;

		m_biasing[i].mean=biasingEqposition[i];
	}
}
