#include <math.h>
#include <iostream>
#include <stdexcept>
#include <limits>
#include <algorithm>
#include <memory>

#include "Utils/include/FileReader.h"
#include "Utils/include/StringParser.h"
#include "Utils/include/Output.h"
#include "Utils/include/Periodicity.h"
#include "Utils/include/SignalHandler.h"

#include "MathTools/include/PhysicalConstants.h"
#include "MathTools/include/MathFunctionOnGrid.h"
#include "MathTools/include/PeriodicTools.h"
#include "MathTools/include/ModelFunction.h"

#include "Common.h"
#include "vFEPCommandManager.h"
#include "MLEFinder.h"
#include "USWindow.h"

#include "USWindowSet.h"

namespace USWindowSet_Impl {
	const unsigned int maxSpline=15;
}
using namespace USWindowSet_Impl;

USWindowSet::USWindowSet(const std::string& metafileName, const int dataFreq)
{
	ReadUSFiles(metafileName, dataFreq);
	SetUpDimensionality();	
	AddImageWindows();
	SetWindowWeight();

	CalculateSplineRange();
	ResetSplineNodes();
}

USWindowSet::~USWindowSet()
{
}


void USWindowSet::ResetSplineNodes(const NDDataSet& newNodes){

	Output& output = Output::Instance();
	std::stringstream ss(std::stringstream::in | std::stringstream::out);
	ss << " Reset the spline nodes." << std::endl;
	output.AddValue("program log", ss);
	
	m_nodePositions = newNodes;
	ResetSplineNodes();
}

void USWindowSet::ResetSplineNodes(){
	
	BuildUpSplineNodes();
	BuildPeriodicCopy();
	BuildUpQuadratureRules();

	vFEPCommandManager& manager = vFEPCommandManager::Instance();
	Output& output = Output::Instance();

	std::stringstream ss(std::stringstream::in | std::stringstream::out);
	for (unsigned i = 0; i < GetDimensionality(); i++){
		ss << "Total spline nodes for dim #" << i + 1 << " are " << m_nodePositions[i].size() << std::endl;
		if (manager.GetKey("v"))
			for (unsigned j = 0; j < m_nodePositions[i].size(); j++) ss << "  Node #" << j + 1 << " : " << m_nodePositions[i][j] << std::endl;
	}
	output.AddValue("program log", ss);

}


bool USWindowSet::AddWindow(USWindow* pWindow, std::string type){
	
	if ( pWindow->TestData()) {

		GetCurrentWindowList().push_back(pWindow);

		Output& output = Output::Instance();
		std::stringstream ss(std::stringstream::in | std::stringstream::out);
		ss << "A ";
		ss << type.c_str();
		ss << " window has been added:\n";
		output.AddValue("program log",ss);
		pWindow->PrintDataInfo();

		return true;
	} else {
		return false;
	}
}

void USWindowSet::GetWindowSetRange(NDData& minR, NDData& maxR){
	
	unsigned dim=GetDimensionality();
	assert(dim>=1);

	minR.resize(dim,1.0e99);
	maxR.resize(dim,-1.0e99);

	const std::vector<DataWindow*>& list=GetCurrentWindowList();

	for (unsigned i=0; i< dim; i++){
		for (unsigned j=0; j< list.size(); j++){
			const GuassianDistribution& biasing=((USWindow*) list[j])->GetBiasing()[i];
			if (minR[i]>biasing.mean) minR[i]=biasing.mean;
			if (maxR[i]<biasing.mean) maxR[i]=biasing.mean;
		}
	}

	return;

}


void USWindowSet::ReadUSFiles(const std::string& metafileName, const int dataFreq){

	vFEPCommandManager& manager= vFEPCommandManager::Instance();

	bool symmetricPair=manager.GetKey("sym");
	bool split=manager.GetKey("nsplit");
	int nSplit=0;
	if (split) {
		nSplit=5;
		manager.GetValue("nsplit", nSplit); 
	}

	FileReader reader(metafileName);
	StringParser parser(reader.GetContains(),"\n");

	const std::vector<std::string>& entries=parser.GetTokens();
	unsigned numberEntries=entries.size();

	Output& output = Output::Instance();
	const Periodicity::PeriodicalConditionSet& c=Periodicity::Instance().GetConditions();
	const double delta=1e-5;

	for (unsigned i=0; i< numberEntries; i++){

		unsigned fileColumnShift=1;
		bool useWeight=false;
		if (manager.GetKey("w")){
			fileColumnShift=2;
			useWeight=true;
		}

		const std::string& entry=entries[i];
		StringParser parser(entry," \t");
		
		const std::vector<std::string>& tokens=parser.GetTokens();
		if (tokens.size()==0) continue;

		const char head=tokens[0][0];
		if (head=='#' || head =='!') {
			std::stringstream ss(std::stringstream::in | std::stringstream::out);
			ss << "Ignored : " << tokens[0].c_str() << "\n";
			output.AddValue("program log",ss);
			continue;
		}

		if (tokens.size()<3) {
			std::string error="**** Problem with the center position and force constant assignment--check your metafile entries ***\n";
			std::runtime_error e(error.c_str());
			throw e;
		} else {

			std::stringstream ss(std::stringstream::in | std::stringstream::out);
			ss << "Reading Window #" << i+1 << "\n";
			output.AddValue("program log",ss);

			if (tokens.size()<4 && useWeight) {
				std::stringstream ss(std::stringstream::in | std::stringstream::out);
				ss << "**** Window weighting option is turned on but not assigned in the metafile. ***\n";
				ss << "Hence the weighting option is turned off now.\n";
				output.AddValue("warning",ss);
				fileColumnShift=1;
				useWeight=false;
			}

			unsigned dimension=(tokens.size()-fileColumnShift)/2;
			const std::string& fileName=tokens[0];
			double weight=1.0;

			NDData position(dimension,0.0), forceConst(dimension,0.0);
			for (unsigned j=0; j<dimension; j++){
				position[j]=atof(tokens[2*j+1].c_str());
				if (c[j].isPeriodic){
					 Wrap(c[j].middle, c[j].period, position[j]);
				}
				forceConst[j]=atof(tokens[2*j+2].c_str());
			}
			if (useWeight){
				double temp=atof(tokens[tokens.size()-1].c_str());
				//weight= (temp>0) ? temp : 1.0;
				weight = temp;
			}

			try {
				USWindow* pWindow (new USWindow (fileName, position, forceConst, weight, dataFreq));
				if ( pWindow->TestData()) {
					if (!split){
						AddWindow(pWindow, "regular");
						if (symmetricPair && ( position[0]!=position[dimension-1]) ){
							USWindow* pairWindow(new USWindow (pWindow->GetRawData(), position, forceConst, weight, true));
							AddWindow(pairWindow, "symmetric-pair");
						}
						
					} else {
						std::vector<USWindow*> newList=USWindow::SplitWindow(pWindow, nSplit);
						if (newList.size()>1){
							for (unsigned i=0; i<newList.size(); i++){
								AddWindow(newList[i], "child");
							}
							std::stringstream ss(std::stringstream::in | std::stringstream::out);
							ss << newList.size()<<" child windows have been added.\n\n";
							output.AddValue("program log",ss);

						} else {
							AddWindow(pWindow, "regular");
						}

					}
				}
				else delete pWindow;
			} catch (std::runtime_error& e){
				std::string s("Warning: ");
				s.append(e.what());
				output.AddValue("warning", s, true);
			}
		}

		SignalHandler& h=SignalHandler::Instance();
		if (h.CheckSignal(2)) {
			std::string error="**** User interuption received.\n";
			std::runtime_error e(error.c_str());
			throw e;
		}

	}

	std::stringstream ss(std::stringstream::in | std::stringstream::out);
	ss << "Done with reading Windows \n\n";
	ss << "Totally number of windows: " << GetCurrentWindowList().size() << "\n\n";

	output.AddValue("program log",ss);
}


void USWindowSet::AddImageWindows(){

	vFEPCommandManager& manager= vFEPCommandManager::Instance();
	Output& output = Output::Instance();
	const Periodicity::PeriodicalConditionSet& c=Periodicity::Instance().GetConditions();
	const double delta=1e-5;

	bool addPB=true;
//	if (manager.GetKey("add_pb")) addPB=true;
	if (manager.GetKey("no_pb")) addPB=false; 

	bool windowAdded=false;

	if (addPB){
		unsigned dimension=((USWindow*)GetCurrentWindowList()[0])->GetDimensionality();
		for (unsigned j=0; j<dimension; j++){
			if (c[j].isPeriodic){

				unsigned currentNumberWindows=GetCurrentWindowList().size();
				for (unsigned i=0; i<currentNumberWindows; i++){
					USWindow* pWindow=(USWindow*) GetCurrentWindowList()[i];
					NDData position(pWindow->GetEqPosition());
					NDData forceConst(pWindow->GetForceConst());	
					double weight = (pWindow->GetWindowWeight());
					
					if (fabs(c[j].max-position[j])<c[j].bufferFactor*c[j].period){
						NDData myPosition(position);
						myPosition[j]-=c[j].period;
						USWindow* pairWindow (new USWindow (pWindow->GetRawData(), myPosition, forceConst, weight, false));
						windowAdded=AddWindow(pairWindow, "periodic-image");
					} 
					if (fabs(position[j]-c[j].min)<c[j].bufferFactor*c[j].period){
						NDData myPosition(position);
						myPosition[j]+=c[j].period;
						USWindow* pairWindow (new USWindow (pWindow->GetRawData(), myPosition, forceConst, weight, false));
						windowAdded=AddWindow(pairWindow, "periodic-image");		
					}
				}// if (c[j].isPeriodic){
			} // for (int i=currentNumberWindows-1; i--; i>=0){
		} // for (unsigned j=0; j<dimension; j++){
	}// if (addPB)

	if (windowAdded){
		std::stringstream ss(std::stringstream::in | std::stringstream::out);
		ss << "Done with adding periodic image windows \n\n";
		ss << "Totally number of windows: " << GetCurrentWindowList().size() << "\n\n";
		output.AddValue("program log",ss);
	}
}


void USWindowSet::CalculateSplineRange(){

	vFEPCommandManager& manager= vFEPCommandManager::Instance();

	unsigned dim=GetDimensionality();	
	
	std::vector<DataWindow*>& list=GetCurrentWindowList();
	unsigned numberWindow=list.size();

	const Periodicity::PeriodicalConditionSet& c=Periodicity::Instance().GetConditions();

	// Calculate the node positions--on the average window data postions and right on the middles
	// For perodic cases, add the minimum and the maximum 
	m_nodePositions.assign(dim, OneDDataSet(0, 0.0));
	
	NDData xmin, xmax, xmin_win, xmax_win; 
	GetDataRange(xmin, xmax);
	double delta=1.0e-5;
	for (unsigned i=0; i<dim; i++){
		if( equal(xmin[i],xmax[i]) ){ xmin[i]-=delta; xmax[i]+=delta; };
	}

	GetWindowSetRange(xmin_win, xmax_win);

	std::vector<std::string> keys(2,"");
	keys[0]="n1"; keys[1]="n2";

	bool autoSpacing[2]={ false, false };
	if (dim>1) autoSpacing[0]=autoSpacing[1]=true;

	int target[2]={-1,-1};

	// Totally empirical determination of the number of spline points
	double f1=pow(1.0*numberWindow, 1.0/dim);
	double f2=2.0-f1*0.07;
	if (f2<1.0) f2=1.0;

	int rr=std::min(maxSpline,static_cast<unsigned int>(f1*f2+1.6));
	int roughNumberSpline[2]={ rr, rr};

	for (unsigned i=0; i<dim; i++){
		target[i]=roughNumberSpline[i];
		if (dim>1) target[i]/=1.2;
		//target[i]=roughNumberSpline[i]/sqrt(double(dim));;
		if (target[i]<5) {
			autoSpacing[i]=true;
			target[i]=5;
		}
		if (c[i].isPeriodic) target[i]*=(1.0+2*c[i].bufferFactor); 
		if (manager.GetKey(keys[i])) {
			autoSpacing[i]=true;
			manager.GetValue(keys[i],target[i]);
			if (target[i]<0 || target[i]>1000) autoSpacing[i]=false;
		}
	}
	
	NDData minInt, maxInt;
	GetIntegrationSize(minInt, maxInt);

	std::vector<double> spline_min=xmin;
	std::vector<double> spline_max=xmax;

	unsigned maxSplineNumber= (dim>1) ? 256 : 50;
	
	for (unsigned i=0; i<dim; i++){

		double totalDataRange=(spline_max[i]-spline_min[i]);
		double miniRange=totalDataRange/(10.0*roughNumberSpline[i]);

		//  Turn on auto spacing if the current dimension is periodic
		if (autoSpacing[i] || c[i].isPeriodic ){  

			double spacing=(spline_max[i]-spline_min[i])/(target[i]*1.0);
			
			// Check if more spline nodes are needed
			if (spacing > minInt[i] && target[i]<10 && !manager.GetKey(keys[i])){
				spacing=minInt[i];
				target[i]=(spline_max[i]-spline_min[i]+0.001)/spacing;
			}

			double starting=spline_min[i];
			if (c[i].isPeriodic){
				int tt= ( (target[i])/(1.0+2*c[i].bufferFactor) );
				spacing=c[i].period/(tt*1.0);
                int shift=(c[i].min-spline_min[i])/spacing;
				starting=c[i].min-shift*spacing;

				double xx=c[i].min - (c[i].max-c[i].min)*c[i].bufferFactor;
				m_nodePositions[i].push_back(std::min(xx,spline_min[i]));
				for (int j=0; j< target[i]; j++){
					double x1=starting+j*spacing;
					m_nodePositions[i].push_back(x1);
				} 

				xx=c[i].max + (c[i].max-c[i].min)*c[i].bufferFactor;
				m_nodePositions[i].push_back(std::max(xx,spline_max[i]));
			} else {
				for (int j=0; j< target[i]+1; j++){
					double x1=starting+j*spacing;
					m_nodePositions[i].push_back(x1);
				} 
			}

		} else {

			m_nodePositions[i].push_back(spline_min[i]);

			double x1=0.0, x2=0.0, xm=0.0;
			for (unsigned j=0; j< numberWindow; j++){
				if (dim==1) x1=((USWindow*) list[j])->GetObserved()[i].mean;
				if (dim==2) x1=((USWindow*) list[j])->GetBiasing()[i].mean;		

				if (x1>spline_min[i] && x1<spline_max[i] ) m_nodePositions[i].push_back(x1);
				
				if ( ( (numberWindow*2)<maxSplineNumber) &&  (j<numberWindow-1)){
					if (dim==1) x2=((USWindow*) list[j+1])->GetObserved()[i].mean;
					if (dim==2) x2=((USWindow*) list[j+1])->GetBiasing()[i].mean;	
					double xm=(x1+x2)/2.0;
					if (xm>spline_min[i] && xm<spline_max[i] ) m_nodePositions[i].push_back(xm);
				}
			}

			m_nodePositions[i].push_back(spline_max[i]);
		}
		std::sort (m_nodePositions[i].begin(), m_nodePositions[i].end());

		// remove nodes which are too close to the neighbors
		for (int j=m_nodePositions[i].size()-1; j>0; j--){
			if ( (m_nodePositions[i][j]-m_nodePositions[i][j-1]) < miniRange )
				m_nodePositions[i].erase(m_nodePositions[i].begin()+j);
		}

	} // for (i=dim...
}

void USWindowSet::BuildUpSplineNodes(){

	unsigned dim=GetDimensionality();	
	assert(dim==m_nodePositions.size());

	vFEPCommandManager& manager = vFEPCommandManager::Instance();
	
	std::vector<DataWindow*>& list=GetCurrentWindowList();
	unsigned numberWindow=list.size();

	//  Assign node positions
	m_splineNodes.clear();
	switch (dim) {
		case 2:
			for (unsigned j=0; j<m_nodePositions[1].size(); j++)
				for (unsigned i=0; i<m_nodePositions[0].size(); i++){
					
					SplineNode node(i, m_nodePositions[0][i], j, m_nodePositions[1][j]);
					
					if (i==0) node.previous[0]=-1e-99;
					else node.previous[0]=m_nodePositions[0][i-1];
					
					if (j==0) node.previous[1]=-1e-99;
					else node.previous[1]=m_nodePositions[1][j-1];
					
					m_splineNodes.push_back(node);
				}
			break;
		case 1: default:
			for (unsigned i=0; i<m_nodePositions[0].size(); i++){
				SplineNode node(i, m_nodePositions[0][i]);
				if (i>0) node.previous[0]=m_nodePositions[0][i-1];
				m_splineNodes.push_back(node);
			}
			break;
	}

	unsigned numberNodes=m_splineNodes.size();

	//  For each window data, calculate the moments <x>, <x^2>, and <x^3> contribution to every nodes  
	for (unsigned i=0; i< numberWindow; i++){
		
		USWindow* pWindow=(USWindow*)list[i];
		if (!pWindow) continue;

		// First identify the range of the possible spline nodes we may use for the data in this window
		DataRangeSet range=pWindow->GetDataRange();
		std::vector<unsigned> lowIndex(dim, 0), highIndex(dim, 0);
		for (unsigned idim=0; idim<dim; idim++) {
			
			unsigned dimNumberNode=m_nodePositions[idim].size();
			highIndex[idim]=m_nodePositions[idim][dimNumberNode-1];

			for (unsigned j=1; j< dimNumberNode; j++){
				if (range[idim].min<=m_nodePositions[idim][j]){
					lowIndex[idim]=j-1;
					break;
				}
			}
			for (int j=dimNumberNode-2; j>=0; j--){
				if (range[idim].max>=m_nodePositions[idim][j]){
					highIndex[idim]=j+1;
					break;
				}
			}	
		}

		// For each data point, calculate the momentum and store the results in 
		// the corresponding spine node
		NDDataSet& positions = pWindow->GetRawData(); 
		unsigned numberData=positions.size();
		
		// uniform weights
		std::vector<double> weight(numberData, pWindow->GetWindowWeight() / numberData);

		bool useDataWeight = (manager.GetKey("dw") || manager.GetKey("dwe") || manager.GetKey("dwb"));
		if (useDataWeight) {
			if (manager.GetKey("dwb")){

				OneDDataSet x0(numberData, 1.0), x1(numberData, 1.0), w0(numberData, 1.0), w1(numberData, 1.0);
				for (unsigned idata = 0; idata < numberData; idata++){
					x0[idata] = positions[idata][0];
					if (dim >= 2)	x1[idata] = positions[idata][1];
				}

				const GuassianDistributionSet& g = pWindow->GetBiasing();
				MathFunctionOnGrid::Gaussian(g[0].mean, g[0].variance, x0, w0, false);
				if (dim >= 2) MathFunctionOnGrid::Gaussian(g[1].mean, g[1].variance, x1, w1, false);
				
				for (unsigned idata = 0; idata < numberData; idata++) weight[idata] = 1.0 / (w0[idata] * w1[idata]);
			}
			else {
				for (unsigned idata = 0; idata < numberData; idata++) weight[idata] = positions[idata][dim];
			}

			double norm = 0.0;
			for (unsigned idata = 0; idata < numberData; idata++) norm += weight[idata];
			for (unsigned idata = 0; idata < numberData; idata++) weight[idata] /= norm;

		}

		for (unsigned idata=0; idata<numberData; idata++){

			std::vector<unsigned> node_index(dim, 0);
			double temp[2]={0.0, 0.0};
			for(unsigned idim=0; idim<dim; idim++){
				double x=positions[idata][idim];
				for (unsigned k=lowIndex[idim]+1; k<=highIndex[idim]; k++){
					if (x<=m_nodePositions[idim][k]){
						node_index[idim]=k;
						temp[idim]=x-m_nodePositions[idim][k-1];
						break;
					}
				}
			}

			// Found the node for the current data point
			unsigned myNode=node_index[0];
			if (dim>=2) myNode+=(node_index[1])*(m_nodePositions[0].size());

			unsigned nx=4, ny=(dim==1 ? 1 :4);

			double xx[4][4]={ {1.0, 1.0, 1.0, 1.0}, 
							  {1.0, 1.0, 1.0, 1.0},	
							  {1.0, 1.0, 1.0, 1.0},
							  {1.0, 1.0, 1.0, 1.0} };

			for (unsigned iy=1; iy<ny; iy++) xx[0][iy]=xx[0][iy-1]*temp[1];
		
			for (unsigned ix=1; ix<nx; ix++){
				xx[ix][0]=xx[ix-1][0]*temp[0];
				for (unsigned iy=1; iy<ny; iy++){
					xx[ix][iy]=xx[ix][iy-1]*temp[1];
				}
			}

			for (unsigned ix=0; ix<nx; ix++)
				for (unsigned iy=0; iy<ny; iy++)
					m_splineNodes[myNode].x[ix][iy]+=xx[ix][iy]*weight[idata];

		} // for (unsigned idata=0; i<numberData
	}

	// Calculate the smallest distance from all data points for each spline node if necessary
	if( manager.GetKey("baye") ){
		NDData xmin, xmax; 
		GetDataRange(xmin, xmax);
		double sx=1.0/(xmax[0]-xmin[0]),sy=1.0/(xmax[1]-xmin[1]);
		double cutOff=( dim==1 ? 1.0/(1.0*numberNodes+1.0) : 1.0/sqrt(1.0*numberNodes+1.0) );
		for (unsigned iNode=0; iNode<numberNodes; iNode++){

			// only those nodes without data points
			//if (equal(m_splineNodes[iNode].x[0][0],0.0)){
				double minDist=1e99;
				for (unsigned i=0; i< numberWindow; i++){
					
					USWindow* pWindow=(USWindow*)list[i];
					if (!pWindow) continue;

					NDDataSet& positions = pWindow->GetRawData(); 
					double weight = pWindow->GetWindowWeight();
					unsigned numberData=positions.size();
					
					double tt=0.0;
					if (dim==1){
						for (unsigned idata=0; idata<numberData; idata++){
							tt=fabs(positions[idata][0]-m_splineNodes[iNode].position[0])*sx;
							if (tt<minDist) minDist=tt;
							if (tt<cutOff) m_splineNodes[iNode].count++;
						}
					} // if (dim==1)
					if (dim==2){
						for (unsigned idata=0; idata<numberData; idata++){
							tt=sqrt((positions[idata][0]-m_splineNodes[iNode].position[0])*(positions[idata][0]-m_splineNodes[iNode].position[0])*sx*sx
								+ (positions[idata][1]-m_splineNodes[iNode].position[1])*(positions[idata][1]-m_splineNodes[iNode].position[1])*sy*sy );
							if (tt<minDist) minDist=tt;
							if (tt<cutOff) m_splineNodes[iNode].count++;
						}
						
					} // if (dim==1)
				} // for (unsigned i=0;
				m_splineNodes[iNode].minDist=minDist;
			//} // if (equal(m_splineNodes[iNode]
		} // for (unsigned iNode=0; i
	} // if( manager.GetKey
	
	return;
}

void USWindowSet::BuildUpQuadratureRules(){

	std::vector<DataWindow*>& list=GetCurrentWindowList();
	unsigned numberWindow=list.size();
	m_ruleSets.clear();

	for (unsigned i=0; i< numberWindow; i++){
		USWindow* pWindow=(USWindow*)list[i];
		if (!pWindow) continue;
		m_ruleSets.push_back(pWindow->GetQuadratureRuleSet());
	}

}


void USWindowSet::BuildPeriodicCopy(){

	const Periodicity::PeriodicalConditionSet& c=Periodicity::Instance().GetConditions();
	unsigned dim=GetDimensionality();

	NDInt boundaryNodes(0,0);
	NDIntSet nodeProperty(dim,NDInt(0,0));

	unsigned numberNodes=m_splineNodes.size();
	m_periodicCopyList.assign(numberNodes, -1);

	for (unsigned i=0; i<numberNodes; i++){
		bool boundary=false;
		NDInt p(2,0);
		for (unsigned j=0; j<dim; j++){
			if (equal(m_splineNodes[i].position[j],c[j].min)) p[j]=-1; 
			else if (equal(m_splineNodes[i].position[j],c[j].max)) p[j]=1;
			if (p[j]!=0) boundary=true;
		}
		if (boundary) {
			boundaryNodes.push_back(i);
			for (unsigned j=0; j<dim; j++){
				nodeProperty[j].push_back(p[j]);
			}
		}
	}
	for (unsigned i=0; i<boundaryNodes.size(); i++){
		int indexi=boundaryNodes[i];
		for (unsigned j=0; j<dim; j++){
			if (nodeProperty[j][i]==-1){
				bool match=false;
				for (unsigned k=i+1; k<boundaryNodes.size(); k++){
					int indexk=boundaryNodes[k];
					if (nodeProperty[j][k]==1) match=true;
					if (dim>1 && match) {
						unsigned jj= (j==0) ? 1 : 0;
						match=equal(m_splineNodes[indexi].position[jj],m_splineNodes[indexk].position[jj]);
					}
					if (match){
						m_periodicCopyList[indexk]=indexi;
						break;
					}
				} // for (unsigned k=i+1; 
			} // if (nodeProperty[j]
		} // for (unsigned j=0;
	} // for (unsigned i=0;
}

void USWindowSet::BuildUpQuadratureRules(const DataRangeSet& dataRange, const GuassianDistributionSet& biasing, QuadratureRuleSet& ruleSet, NDData& intSize){

	vFEPCommandManager& manager = vFEPCommandManager::Instance();

	double temperature = 300;
	if (manager.GetKey("temp")) manager.GetValue("temp", temperature);
		
	double beta = 1.0 / (PhysicalConstants::GasConstantInKcalPerMole*temperature);

	bool directFEP = manager.GetKey("directfep");
	bool biasingDataWeighting = manager.GetKey("dwb");

	unsigned numberDimension = dataRange.size();

	ruleSet.clear();

	intSize.clear();
	intSize.resize(numberDimension, 0.0);

	for (unsigned i = 0; i<numberDimension; i++) {

		intSize[i] = (dataRange[i].max - dataRange[i].min) * 20;  //  an arbitrary large number
		bool useGaussLegendre = manager.GetKey("qgl");
		bool outsideRange = ((biasing[i].mean>dataRange[i].max) || (biasing[i].mean<dataRange[i].min));

		bool noBiasing = (equal(biasing[i].variance, 0) || biasingDataWeighting);

		if (noBiasing || directFEP) useGaussLegendre = true;

		//if ((m_observed[i].variance/m_biasing[i].variance)>3.0) useGaussLegendre=true;
		//if (outsideRange) useGaussLegendre=true;

		unsigned numberQuadrature = (useGaussLegendre) ? numberQuadradureWeightPoints / 2 : numberQuadradureWeightPoints;
		if (manager.GetKey("nq")) manager.GetValue("nq", numberQuadrature);
		if (outsideRange) numberQuadrature *= 2;

		if (!useGaussLegendre) {
			//Using biasing potential as the weight for GaussHermit rules
			ruleSet.push_back(QuadratureRule(numberQuadrature, biasing[i]));
			for (int inode = numberQuadrature - 1; inode >= 0; inode--){
				double t1 = ruleSet[i].rule[inode].position - dataRange[i].min;
				double t2 = ruleSet[i].rule[inode].position - dataRange[i].max;
				if (t1<0 || t2>0) ruleSet[i].rule.erase(ruleSet[i].rule.begin() + inode);
			}

		}
		else {

			double fullRange = (dataRange[i].max - dataRange[i].min);

			unsigned nSegment = 12;
			if (manager.GetKey("nseg")) manager.GetValue("nseg", nSegment);
			numberQuadrature -= (numberQuadrature % nSegment); // make the total numberQuadrature divisible by nSegment	
			if (numberQuadrature<nSegment * 4) numberQuadrature = nSegment * 4; // at least four points for each segment

			intSize[i] = fullRange / (nSegment*1.0);

			unsigned segNumberQuadrature = numberQuadrature / nSegment;
			QuadratureRule currentRule;

			for (unsigned iSeg = 0; iSeg<nSegment; iSeg++){

				double range = fullRange / (nSegment*2.0);
				double refMean = range*(2.0*iSeg + 1.0) + dataRange[i].min;

				unsigned nodeStart = iSeg*segNumberQuadrature;
				unsigned nodeEnd = (iSeg + 1)*segNumberQuadrature - 1;

				//currentRule+=QuadratureRule(segNumberQuadrature, refMean, range, 1.0/(nSegment*1.0));
				currentRule += QuadratureRule(segNumberQuadrature, refMean, range);

				if (!noBiasing){

					std::vector<double> p(segNumberQuadrature, 0.0), biasingPotential(segNumberQuadrature, 0.0);
					for (unsigned inode = nodeStart; inode <= nodeEnd; inode++) p[inode - nodeStart] = currentRule.rule[inode].position;

					if (directFEP){
						for (unsigned inode = nodeStart; inode <= nodeEnd; inode++)
							biasingPotential[inode - nodeStart] = -biasing[i].variance*beta*beta*p[inode - nodeStart];
					}
					else {
						MathFunctionOnGrid::Quadratic(0.5 / biasing[i].variance, biasing[i].mean, p, biasingPotential);
					}
					for (unsigned inode = nodeStart; inode <= nodeEnd; inode++)	currentRule.rule[inode].logWeight -= biasingPotential[inode - nodeStart];
				}

			}
			ruleSet.push_back(currentRule);
		}

	} // for (unsigned i=0; i<numberDimension;

}